import obspy
import numpy as np

def aligned_sac_datas(tr1,tr2,scc_dt,tb,te,marker='t0'):
    """
    Parameters:
       tr1,tr2: obspy trace
        scc_dt: time difference measured by cross-correlation, t1-t2
            tb: Waveform cut time before the S arrival
            te: Waveform cut time after the S arrival
        marker: 'a' for P arrival time, 't0' for S arrival time
    """
    tr1.detrend("linear")
    tr1.detrend("constant")
    tr2.detrend("linear")
    tr2.detrend("constant")
    
    t01 = getattr(tr1.stats.sac,marker)
    t02 = getattr(tr2.stats.sac,marker)
    b1 = tr1.stats.sac.b
    b2 = tr2.stats.sac.b
    trim_b1 = tr1.stats.starttime+(t01-b1)+scc_dt+tb
    trim_e1 = trim_b1+(te-tb)

    trim_b2 = tr2.stats.starttime+(t01-b2)+tb
    trim_e2 = trim_b2+(te-tb)

    tr1.trim(starttime = trim_b1, endtime = trim_e1)
    tr2.trim(starttime = trim_b2, endtime = trim_e2)
    
    tr1.detrend("constant")
    tr1.detrend("linear")
    tr2.detrend("constant")
    tr2.detrend("linear")

    return tr1.data,tr2.data



import warnings
from seisloc.hypoinv import load_y2000
import numpy as np
import os
import re
import matplotlib.pyplot as plt

def gen_abs_file(arcfile,outfile="Input_Files/absolute.dat",pweight=1,sweight=0.5):
    pweight = 1
    sweight = 0.5
    arcphs = load_y2000(arcfile)
    with open("Input_Files/absolute.dat",'w') as f:
        for key in arcphs.keys():
            evid = arcphs[key]["evid"]
            f.write(f"#             {format(str(evid),'>5s')}\n")
            phases = arcphs[key]["phase"]
            for phase in phases:
                net,sta,phs,abstime = phase
                f.write(f"{format(sta,'<5s')}")
                f.write("      ")
                f.write(f"{format(abstime,'>5.2f')}")
                f.write("       ")
                if phs == "P":
                    f.write(format(pweight,'3.1f'))
                elif phs == "S":
                    f.write(format(sweight,'3.1f'))
                f.write("   ")
                f.write(phs)
                f.write("\n")


def prepMOD(head,lon_list,lat_list,dep_list,vel_list,poisson_list):
    """
    Output MOD file for the tomoDD based on information provided
    Parameters:
    head: bld,nx,ny,nz. bld:resolution; nx/ny/nz: nodes for lon/lat/dep
    vel_list: P wave velocity list
    poisson_list: possion ratio of each layer
    len(lon_list)=nx; len(lat_list)=ny; len(dep_list)=nz;
    len(vel_list)==nz;len(poisson_list)==nz
    """
    f = open("MOD",'w')
    bld = head[0];
    nx = head[1];
    ny = head[2];
    nz = head[3];
    if nx != len(lon_list):
        raise Exception("Wrong longitude list length")
    if ny != len(lat_list):
        raise Exception("Wrong latitude list length")
    if nz != len(dep_list):
        raise Exception("Wrong depth list length")
    if nz != len(vel_list):
        raise Exception("Wrong velocity list length")
    if nz != len(poisson_list):
        raise Exception("Wrong poisson list length")

    if len(lon_list) != len(set(lon_list)):
        raise Exception("Duplicated values in longitude list.")
    if len(lat_list) != len(set(lat_list)):
        raise Exception("Duplicated values in latitude list.")
    if len(dep_list) != len(set(dep_list)):
        raise Exception("Duplicated values in depth list.")

    for i in range(len(lon_list)-1):
        if lon_list[i]>lon_list[i+1]:
            warnings.warn(f"lon_list[{i}]>lon_list[{i+1}]")
    for i in range(len(lat_list)-1):
        if lat_list[i]>lat_list[i+1]:
            warnings.warn(f"lat_list[{i}]>lat_list[{i+1}]")
    for i in range(len(dep_list)-1):
        if dep_list[i]>dep_list[i+1]:
            warnings.warn(f"dep_list[{i}]>dep_list[{i+1}]")

    f.write(f"{bld} {nx} {ny} {nz}\n")
    for i in range(len(lon_list)):
        f.write(str(lon_list[i]))
        if i != len(lon_list):
            f.write(" ")
    f.write("\n")
    for i in range(len(lat_list)):
        f.write(str(lat_list[i]))
        if i != len(lat_list):
            f.write(" ")
    f.write("\n")
    for i in range(len(dep_list)):
        f.write(str(dep_list[i]))
        if i != len(dep_list):
            f.write(" ")
    f.write("\n")
        
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                f.write(format(vel_list[k],'5.3f'))
                if i != (nx-1):
                    f.write(" ")
            f.write("\n")
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                f.write(format(poisson_list[k],'5.3f'))
                if i != (nx-1):
                    f.write(" ")
            f.write("\n")
    f.close()

def load_xyz(locfile="tomoDD.reloc"):
    mdat = np.loadtxt(locfile);
    cusp = mdat[:,0];lon=mdat[:,2];lat=mdat[:,1];mag=mdat[:,16];
    x = lon; y=lat; z=mdat[:,3]
    return x,y,z

def load_MOD(MODfile):
    MOD = []
    with open(MODfile,'r') as f:
        for line in f:
            MOD.append(line.rstrip())
    _,_nx,_ny,_nz=MOD[0].split()
    global nx,ny,nz
    nx = int(_nx); ny = int(_ny); nz = int(_nz);
    X = np.zeros((1,nx))
    _X = MOD[1].split() # Second line is x list 
    for i in range(len(_X)):
        X[0,i]=float(_X[i])
    Y = np.zeros((1,ny))
    _Y = MOD[2].split() # Third line is y list
    for i in range(len(_Y)):
        Y[0,i]=float(_Y[i])
    Z = np.zeros((1,nz))
    _Z = MOD[3].split() # Forth line is z list
    for i in range(len(_Z)):
        Z[0,i] = float(_Z[i])
    VpVs = np.loadtxt(MODfile,skiprows=4)    
    VEL_P = np.zeros((nx,ny,nz))
    VEL_S = np.zeros((nx,ny,nz))  
    
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                VEL_P[i,j,k] = VpVs[k*ny+j,i]
                VEL_S[i,j,k] = VpVs[k*ny+j,i]/VpVs[nz*ny+k*ny+j,i]
    return X,Y,Z, VEL_P,VEL_S

class MOD():
    def __init__(self,MODfile="MOD"):
        self.X,self.Y,self.Z,self.Vp, self.Vs = load_MOD(MODfile)
        self.nx = self.X.shape[-1]
        self.ny = self.Y.shape[-1]
        self.nz = self.Z.shape[-1]

def load_tomo_vel(nx,ny,nz,Vpfile="Vp_model.dat",Vsfile="Vs_model.dat"):
    Vp = np.loadtxt(Vpfile)
    Vs = np.loadtxt(Vsfile)
    VEL_P = np.zeros((nx,ny,nz))
    VEL_S = np.zeros((nx,ny,nz))
    POS_RATIO = np.zeros((nx,ny,nz))
    for k in range(nz):
        for j in range(ny):
            for i in range(nx):
                VEL_P[i,j,k] = Vp[k*ny+j,i]
                VEL_S[i,j,k] = Vs[k*ny+j,i]
                POS_RATIO[i,j,k] = Vp[k*ny+j,i]/Vs[k*ny+j,i]
    
    return VEL_P,VEL_S,POS_RATIO

def prep_D3MOD(MODfile="MOD",Vpfile="Vp_model.dat",Vsfile="Vs_model.dat"):
    assert os.path.exists("D3MOD") == False
    mod = MOD(MODfile=MODfile)
    VEL_P,VEL_S,POS_RATIO = load_velocity(mod.nx,mod.ny,mod.nz,Vpfile=Vpfile,Vsfile=Vsfile)
    with open("D3MOD",'w') as f:
        for line in mod.lines[:4]:
            f.write(line+"\n")
        for k in range(mod.nz):
            for j in range(mod.ny):
                for i in range(mod.nx):
                    f.write(format(VEL_P[i,j,k],'5.3f'))
                    if i != (mod.nx-1):
                        f.write(" ")
                f.write("\n")

        for k in range(mod.nz):
            for j in range(mod.ny):
                for i in range(mod.nx):
                    f.write(format(POS_RATIO[i,j,k],'5.3f'))
                    if i != (mod.nx-1):
                        f.write(" ")
                f.write("\n")

def tomo_event_sel(evid_list=[],event_dat="event.dat",event_sel="event.sel"):
    '''
    select events in the "event.dat" file and output them into
    the "event.sel" file by the event ID list provided
    '''
    content = []
    # Read in data
    with open(event_dat,'r') as f:
        for line in f:
            line = line.rstrip()
            evid = int(line[-8:-1])
            if evid in evid_list:
                content.append(line)
    f.close()

    # Output into target file
    with open(event_sel,'w') as f:
        for line in content:
            f.write(line+"\n")
    f.close()

def extract_log_info(damp_log = "tomoDD-SE.log"):
    smooth_damp_list = []    # smooth and damp line
    abs_ccs = []
    wt_ccs = []
    abs_cts = []
    wt_cts = []
    with open(damp_log,'rb') as f:
        for line in f:
            if line[:7]==b" smooth":
                tmp = re.split(":",line.decode("ascii"))[1]
                _,_smooth,_damp,_xnorm,_xnorm_vel,_rnorm,_rnorm_wt = re.split(" +",tmp.rstrip())
                smooth_damp_list.append([float(_smooth),float(_damp),float(_xnorm),float(_xnorm_vel),
                                       float(_rnorm),float(_rnorm_wt)])
            if line[:16]==b" absolute cc rms":
                tmp = re.split("=",line.decode("ascii"))[1]
                _value = re.split(" +",tmp.rstrip())[1]
                abs_ccs.append(float(_value))
            if line[:16]==b" weighted cc rms":
                tmp = re.split("=",line.decode("ascii"))[1]
                _value = re.split(" +",tmp.rstrip())[1]
                wt_ccs.append(float(_value))
            if line[:16]==b" absolute ct rms":
                tmp = re.split("=",line.decode("ascii"))[1]
                _value = re.split(" +",tmp.rstrip())[1]
                abs_cts.append(float(_value))
            if line[:16]==b" weighted ct rms":
                tmp = re.split("=",line.decode("ascii"))[1]
                _value = re.split(" +",tmp.rstrip())[1]
                wt_cts.append(float(_value))
    log_info_dict = {}
    log_info_dict["smooth_damp_list"] = np.array(smooth_damp_list)
    log_info_dict["abs_ccs"] = abs_ccs
    log_info_dict["wt_ccs"] = wt_ccs
    log_info_dict["abs_cts"] = abs_cts
    log_info_dict["wt_cts"] = wt_cts
    return log_info_dict

def log_info_plot(log_info_dict):
    fig,axs = plt.subplots(3,3,figsize=(12,8))
    niter = log_info_dict["smooth_damp_list"].shape[0]
    axs[0,0].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,2])
    axs[0,0].set_ylabel("xnorm")
    axs[0,1].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,3])
    axs[0,2].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,3]/log_info_dict["smooth_damp_list"][:,2])
    axs[0,2].set_ylabel("Ratio")
    axs[0,1].set_ylabel("xnorm_vel")
    axs[1,0].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,4])
    axs[1,0].set_ylabel("rnorm")
    axs[1,1].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,5])
    axs[1,1].set_ylabel("rnorm_wt")
    axs[1,2].plot(np.arange(niter)+1,log_info_dict["smooth_damp_list"][:,5]/log_info_dict["smooth_damp_list"][:,4])
    axs[1,2].set_ylabel("Ratio")
    miter = len(log_info_dict["abs_cts"])
    axs[2,0].plot(np.arange(miter)+1,log_info_dict["abs_cts"],'k-',label="abs_cts")
    axs[2,0].plot(np.arange(miter)+1,log_info_dict["wt_cts"],'b-',label="wt_cts")
    axs[2,0].legend()
    axs[2,1].plot(np.arange(miter)+1,log_info_dict["abs_ccs"],'k-',label="abs_ccs")
    axs[2,1].plot(np.arange(miter)+1,log_info_dict["wt_ccs"],'b-',label="wt_ccs")
    axs[2,1].legend()
    plt.tight_layout()
    plt.show()



#!/usr/bin/env python
# coding: utf-8

import obspy
import glob
import os
import re
from obspy import UTCDateTime
import warnings
import numpy as np
import multiprocessing as mp
import time
from tqdm import tqdm
from seisloc.sta import load_sta
warnings.filterwarnings("ignore")

def pos_write(o_folder,sta_dict):
    '''
    position write of station information
    '''
    sac_list=glob.glob(o_folder+"/"+"*.SAC")
    inp_status = False
    st = obspy.read(sac_list[0])
    net = st[0].stats.network
    sta = st[0].stats.station
    for sac in sac_list:
        st=obspy.read(sac)
        st[0].stats.sac.stla=sta_dict[net][sta][0]
        st[0].stats.sac.stlo=sta_dict[net][sta][1]
        st[0].stats.sac.stel=sta_dict[net][sta][2]
        st[0].write(sac)
        
def time_file_list(i_path,folder):
    '''
    Get the starttime and file path in sequence, append the last
    end time to the time list.
    '''
    file_list=os.listdir(os.path.join(i_path,folder))
    data_path_list = []
    t_point_list = []
    max_endtime = UTCDateTime(1900,1,1)
    for file in file_list:
        file_path = os.path.join(i_path,folder,file)
        try:
            st=obspy.read(file_path,headonly=True)
            starttime = st[0].stats.starttime
            t_point_list.append(starttime)
            endtime = st[0].stats.endtime
            if endtime>max_endtime:
                max_endtime = endtime
            data_path_list.append(file_path)
        except:    # not seismic file
            continue     
    tmp = np.array(t_point_list)
    data_path_list = np.array(data_path_list)
    k = tmp.argsort()
    data_path_list = data_path_list[k]
    t_point_list.sort()
    t_point_list.append(max_endtime)
    return t_point_list,data_path_list


def day_split(i_path,o_path,folder,sta_dict,format,shift_hour=0,cut_data=True):
        # get the ordered time list and corresponding file path 
        t_point_list,data_path_list = time_file_list(i_path,folder)
        
        if len(data_path_list)==0:
            return
 
        st = obspy.read(data_path_list[0],headonly=True)
        net = st[0].stats.network
        sta = st[0].stats.station

        o_folder = os.path.join(o_path,sta)
        if not os.path.exists(o_folder):
            os.mkdir(o_folder)

        for chn in ["BHE","BHN","BHZ"]:

            tts=t_point_list[0] #total start time is the start time of the first file 
            tte=t_point_list[-1]#total end time is the end time of the last file
            #start to get the trim point to form day_sac files
            trim_point_list=[]
            trim_point_list.append(tts)#the first point is the start time of the first file
            #information of the first day
            f_year=t_point_list[0].year
            f_month=t_point_list[0].month
            f_day=t_point_list[0].day
            #the second point should be the start time of the second day
            trim_node=UTCDateTime(f_year,f_month,f_day)+24*60*60+shift_hour
            #if the second point is less than the total end time, add it into list and move to next day
            while trim_node < tte:
                trim_point_list.append(trim_node)
                trim_node+=24*60*60
            #add the total time end to the last
            trim_point_list.append(tte)
            # write into txt file available days
            f = open(os.path.join(o_folder,'availdays.txt'),'w')
            tmp_list = []
            for time in trim_point_list:
                if time.julday not in tmp_list:
                    f.write(f"{time.julday}\n")
                    tmp_list.append(time.julday)
            f.close()
            if cut_data == False:
                return                  #only generate availdays.txt
            #for each time segment, need to check the corresponding files then load, merge and trim.
            for i in range(len(trim_point_list)-1):
                trim_s=trim_point_list[i] #trim start time
                trim_e=trim_point_list[i+1]#trim end time
                #get the index of trim_s and trim_e file
                for j in range(len(t_point_list)-1):
                    if trim_s>=t_point_list[j] and trim_s<t_point_list[j+1]:
                        count_s=j
                for k in range(len(t_point_list)-1):
                    if trim_e>=t_point_list[k] and trim_s<t_point_list[k+1]:
                        count_e=k
                #read in the coresponding file
                st=obspy.read(data_path_list[count_s])
                #read more file to include the trim_s and trim_e time
                while count_s < count_e:
                    count_s+=1
                    st+=obspy.read(data_path_list[count_s])
                #merge data, using try except to avoid mistake len(st)==1
                try:
                    st.merge(method=1,fill_value="interpolate")
                except:
                    pass
                #trim the data
                st.trim(starttime=trim_s,endtime=trim_e)
                #save the data
                for tr in st:
                    chn=tr.stats.channel
                    if format.upper()=="SAC":
                        f_name=net+"."+sta+"."+chn+"__"+\
                               trim_s.strftime("%Y%m%dT%H%M%SZ")+"__"+\
                               trim_e.strftime("%Y%m%dT%H%M%SZ")+".SAC"
                    if format.upper()=="MSEED":
                        f_name=net+"."+sta+"."+chn+"__"+\
                               trim_s.strftime("%Y%m%dT%H%M%SZ")+"__"+\
                               trim_e.strftime("%Y%m%dT%H%M%SZ")+".mseed"
                    tr.write(o_folder+"/"+f_name,format=format)
                    
        if format=="SAC":
            pos_write(o_folder,sta_dict)
            
            
def mp_day_split(i_path,o_path,sta_file,format="mseed",shift_hour=0,parallel=True,cut_data=True):
    '''
    This function reads in the data from QS5A devices and split them by days
    Parameters:
    i_path: base folder containing all the source files. base_folder/sta_folder/seismic_files
    o_path: output path. output_path/sta_name/seismic_files.
    sta_file: station files containing the longitude and latitude information
    format: "SAC" or "MSEED"
    shift_hour: Shift the trimming time with reference to hour 0. The main purpose is the adjust time zone issue
    '''
    # First load in the station file
    print("# Loading station files ...")
    sta_dict = load_sta(sta_file)

    dir_list = os.listdir(i_path) # Get dir list before create below folder 
    if not os.path.exists(o_path):
        os.mkdir(o_path)
    folder_list=[]
    for item in dir_list:
        if os.path.isdir(os.path.join(i_path,item)) and item[0] !="\.":
            folder_list.append(item)
    print("# Folders to be processed: ",len(folder_list))
    
    # Now check whether all the stations 
    status = True
    false_folder = []
    print("# Check the network and station name...")
    for folder in tqdm(folder_list):
        file_list=os.listdir(os.path.join(i_path,folder))
        for file in file_list:
            try:
                st = obspy.read(os.path.join(i_path,folder,file),headonly=True)
                net = st[0].stats.network
                sta = st[0].stats.station
                if net not in sta_dict:
                    false_folder.append([folder,net,sta])
                else:
                    try:
                        if sta not in sta_dict[net]:
                            false_folder.append([folder,net,sta])
                        else:
                            pass
                    except:
                        false_folder.append([folder,net,sta])
                break
            except:
                continue
    if len(false_folder)>0:
        status = False
        for record in false_folder:
            print(f"Error net or sta in folder: {record[0]} net:{record[1]} sta:{record[2]}")
        
    if status == True and parallel==True:
        print("# Now multiprocessing...")
        cores = int(mp.cpu_count()/2)
        tasks = []
        for folder in folder_list:
            tasks.append([i_path,o_path,folder,sta_dict,format,shift_hour,cut_data])
        pool = mp.Pool(processes=cores)
        rs = pool.starmap_async(day_split,tasks,chunksize=1)
        pool.close()
        while True:
            remaining = rs._number_left
            print(f"Finished {len(tasks)-remaining}/{len(tasks)}",end='\r')
            if(rs.ready()):
                break
            time.sleep(0.5)
    if status == True and parallel==False:
        for folder in folder_list:
            day_split(i_path,o_path,folder,sta_dict,format,shift_hour,cut_data=cut_data)

if __name__=="__main__":
    """
    This function reads in the data from QS5A devices and split them by days
    Parameters:
    i_path: base folder containing all the source files. base_folder/sta_folder/seismic_files
    o_path: output path. output_path/sta_name/seismic_files.
    sta_file: station files containing the longitude and latitude information
    """
    i_path = "./raw_data"
    o_path = './day_data'
    sta_file = './sta.txt'
    
    mp_day_split(i_path,o_path,sta_file)



import matplotlib.pyplot as plt
import glob
import os
import re
import numpy as np
import pickle
from tqdm import tqdm
from numba import jit

def xcorr(data1,data2,max_shift_num):
    len1 = len(data1)
    len2 = len(data2)
    min_len = min(len1,len2)
    cross_list = []
    for shift_num in np.arange(-max_shift_num,max_shift_num+1,1):
        if shift_num<0:
            correlate_value = np.correlate(data1[:min_len+shift_num],data2[-shift_num:min_len])
            cross_list.append(correlate_value.ravel())
        else:
            correlate_value = np.correlate(data2[:min_len-shift_num],data1[shift_num:min_len])
            cross_list.append(correlate_value.ravel())
    cross_list = np.array(cross_list)
    return cross_list.ravel()

@jit(nopython=True)
def xcorr_jit(data1,data2,max_shift_num):
    len1 = len(data1)
    len2 = len(data2)
    min_len = min(len1,len2)
    cross_list = np.zeros(2*max_shift_num+1)
    for i,shift_num in enumerate(np.arange(-max_shift_num,max_shift_num+1,1)):
        if shift_num<0:
            cross_list[i] = np.dot(data1[:min_len+shift_num],data2[-shift_num:min_len])
        else:
            cross_list[i] = np.dot(data2[:min_len-shift_num],data1[shift_num:min_len])

    return cross_list

def plot_corr(ax,sta_pair,base="CFs_Result",xlim=[-10,10],cmap="PiYG"):
    """
    Plot the ambient noise results of one station pair on ax delivered.
    
    Parameters
             ax: ax of matplot.pyplot delivered
       sta_pair: station pair of two stations
           base: base path for the cross-correlation result
           xlim: lag time range for plot in seconds
    """
    file_list = glob.glob(os.path.join(base,sta_pair,"*.pkl"))
    file_list.sort()
    pkls = []
    for file in file_list:
        f = open(file,'rb')
        pkls.append(pickle.load(f))
    if len(pkls) == 0:
        return False
    data = np.zeros((len(pkls),len(pkls[0]["NCF"])))
    for i in range(len(pkls)):
        data[i,:] = pkls[i]["NCF"].ravel()
    [xi,yi]=np.meshgrid(pkls[0]['CFtime'],np.arange(0,len(pkls)+1,1))
    ax.pcolormesh(xi,yi,data,cmap="PiYG",vmin=-1,vmax=1)
    ax.vlines(0,0,len(pkls))
    ax.set_ylim([len(pkls),0])
    ax.set_xlim(xlim)
    ax.set_xlabel("Lag Time (s)")
    ax.set_ylabel("Day")
    ax.set_title(sta_pair)
    return True

def plot_pairs_corr(base="CFs_Result",
                   xlim=[-10,10],
                   figsize=(8,4),
                   exclude_folders=["Logs","Plot"],cmap="PiYG"):
    """
    Plot ambient noise results of each pair under the base path

    Parameters
               base: base path for the cross-correlation result
               xlim: lag time range for plot in seconds
            figsize: figure size for plot
    exclude_folders: exclude folders as they are not station pair folders
    """
    for item in os.listdir(base):
        item_path = os.path.join(base,item)
        if os.path.isdir(item_path) and item not in exclude_folders:
            sta_pair = item
            fig,ax = plt.subplots(1,1,figsize=figsize)
            status = plot_corr(ax,sta_pair,base,xlim=xlim,cmap=cmap)
            if status == True:
                fig.savefig(os.path.join(base,sta_pair,f"{sta_pair}.pdf"))


def plot_stas_corr(sta_list=[],
                   base="CFs_Result",
                   xlim=[-5,5],
                   ax_xsize=4,
                   ax_ysize=4,
                   cmap="PiYG"):
    """
    For each sta in sta_list, extract all the pairs contain this sta and plot
    in one figure.

    Parameters
      sta_list: stations to be processed,empty means all stations under base
          base: the folder cotaining cross-correlation results
          xlim: lag time range for plot in seconds
      ax_xsize: x size for each ax
      ax_ysize: y size for each ax
    """
    if len(sta_list)==0:                     # process all stations
        for item in os.listdir(base):
            if os.path.isdir(os.path.join(base,item)) and \
               len(item.split("_"))==2:
                sta1,sta2 = item.split("_")
                if sta1 not in sta_list:
                    sta_list.append(sta1)
                if sta2 not in sta_list:
                    sta_list.append(sta2)
    for sta in tqdm(sta_list):
        process_list = []
        for sta_pair in os.listdir(base):
            if sta in sta_pair:
                process_list.append(sta_pair)
        k = len(process_list)
        if k>0:
            if k == 1:
                fig,ax = plt.subplots(1,1,figsize=(ax_xsize,ax_ysize))
                plot_corr(ax,process_list[0],base)
                return
            elif k <=4:
                fig,axs = plt.subplots(1,k,figsize=(4*ax_xsize,ax_ysize))
            else:
                if (k/4-int(k/4))==0:
                    rows =int(k/4)
                else:
                    rows = int(k/4)+1
                columns = 4
                fig,axs = plt.subplots(rows,columns,
                                       figsize=(columns*ax_xsize,rows*ax_ysize)
                                       )
            axs_list = axs.ravel()
            for i in range(len(process_list)):
                plot_corr(axs_list[i],process_list[i],base=base,xlim=xlim,cmap=cmap)
            plt.tight_layout()
            plot_folder = os.path.join(base,"Plot") 
            if not os.path.exists(plot_folder):
                os.mkdir(plot_folder)
            plt.savefig(os.path.join(base,"Plot",f"{sta}.png"))



#----------------------------------------------------------------------------
# coding: utf-8
# Author: ZI,Jinping
# History:
#     2021-04-06 Initial coding
###################################

import os
import numpy as np
from math import radians,cos,acos,sin,asin,sqrt,ceil,pi,floor
import obspy
from obspy import Stream
import glob
import re
from obspy import UTCDateTime
import pandas as pd
import matplotlib.pyplot as plt
import warnings
from obspy.geodetics import gps2dist_azimuth
from tqdm import tqdm
import multiprocessing as mp
import random
import shutil
import subprocess
import time
import pickle
from seisloc.geometry import in_rectangle,loc_by_width


def invmod2vel(out_file,vp_file,vs_file="",ps_ratio=1.73,vpdamp=1,vsdamp=1):
    """
    Convert hypoinverse velocity model to the velest velocity model

    Parameters:
       out_file: the output velest velocity model file
        vp_file: input hypoinverse P velocity file
        vs_file: input hypoinverse S velocity file, if vs_file == "",
                   the output S velocity will be calculated based on P
                   velocity and ps_ratio
       ps_ratio: used when vs_file==""
         vpdamp: damping factor in the output P velocity model
         vsdamp: damping factor in the output S velocity model
    """
    vp_vels = []
    vp_lays = []
    vs_vels = []
    vs_lays = []
    with open(vp_file,'r') as f:
        cont = f.readlines()
    for i in range(1,len(cont)):                 # Start from the second line
        vp_vel,vp_lay = map(float,cont[i].split())
        vp_vels.append(vp_vel)
        vp_lays.append(vp_lay)
    if vs_file != "":
        with open(vs_file,'r') as f:
            cont = f.readlines()
        for i in range(1,len(cont)):             # Start from the second line
            vs_vel,vs_lay = map(float,cont[i].split())
            vs_vels.append(vp_vel)
            vs_lays.append(vp_lay)
    else:
        vs_lays = vp_lays.copy()
        for vp_vel in vp_vels:
            vs_vels.append(vp_vel/ps_ratio)

    f = open(out_file,'w')
    f.write("Velocity model from HYPOINVERSE\n")
    f.write(str(len(vp_vels))+"\n")
    for i in range(len(vp_vels)):
        f.write(format(vp_vels[i],'5.2f'))
        f.write("     ")
        f.write(format(vp_lays[i],'7.2f'))
        f.write("  ")
        f.write(format(vpdamp,'7.3f'))
        f.write("\n")
    f.write(str(len(vs_vels))+"\n")
    for i in range(len(vs_vels)):
        f.write(format(vs_vels[i],'5.2f'))
        f.write("     ")
        f.write(format(vs_lays[i],'7.2f'))
        f.write("  ")
        f.write(format(vpdamp,'7.3f'))
        f.write("\n")

def phs_add_mag(phs_file,mag_file):
    """
    Add magnitude information to the phs file. If no magnitude, set it to -9
    """
    out_file = phs_file+".mag"

    # load in magnitude information
    mag_dict = {}
    with open(mag_file,'r') as f:
        for line in f:
            line = line.strip()
            _evid,_mag = re.split(" +",line)
            evid = int(_evid)
            mag = float(_mag)
            mag_dict[evid] = mag
    
    output_lines = []
    evid_line_idxs = []
    evid_list = []
    i = 0
    tmp_lines = []

    with open(phs_file,'r') as f:
        for line in f:
            line = line.rstrip()
            if line[:5] != "     ":
                tmp_lines.append(line)
            if line[:5] == "     ":
                tmp_lines.append(line)
                evid = int(re.split(" +",line)[1])
                try:
                    e_mag = mag_dict[evid]
                    if e_mag < 0:
                        e_mag = 0
                except:
                    e_mag = 0
                eve_line = tmp_lines[0]
                if len(eve_line) < 126:
                    eve_line = eve_line[:36]+str(int(e_mag*100)).zfill(3)+eve_line[39:]
                output_lines.append(eve_line)  # append the event line
                for line in tmp_lines[1:]:
                    output_lines.append(line)
                tmp_lines = []                 # empty the temporary line
    f.close()

    with open(phs_file+".mag","w") as f:
        for line in output_lines:
            f.write(line+"\n")


def load_y2000(y2000_file,print_line=False):
    """
    If print_line is true, each phase line will be printed out
    """
    phs_cont = []
    with open(y2000_file,"r") as f1:
        for line in f1:
            phs_cont.append(line.rstrip())
    f1.close()
    phs_dict = {}
    event_count = 0

    print(">>> Loading phases ... ")
    for line in tqdm(phs_cont):
        if print_line:
            print(line)
        f_para = line[0:2]     # first two characters as first parameter(f_para)
        if re.match("\d+",f_para):    # event line
            event_count += 1
            _yr=line[0:4];_mo=line[4:6];_day=line[6:8]
            _hr=line[8:10];_minute=line[10:12];
            yr = int(_yr); mo = int(_mo); day=int(_day); hr=int(_hr);minute=int(_minute);
            _seconds=line[12:14]+"."+line[14:16]
            lat_deg = int(line[16:18]) # short of latitude degree
            try:
                lat_min_int = int(line[19:21]) 
            except:
                lat_min_int = 0
            try:
                lat_min_decimal = int(line[21:23])*0.01
            except:
                lat_min_decimal = 0
            lat_min = lat_min_int + lat_min_decimal
            evla = lat_deg + lat_min/60

            lon_deg = int(line[23:26]) # short of longitude degree
            try:
                lon_min_int = int(line[27:29])
            except:
                lon_min_int = 0
            try:
                lon_min_decimal = int(line[29:31])*0.01
            except:
                lon_min_decimal = 0
            lon_min = lon_min_int + lon_min_decimal  # short of longitude minute    
            evlo = lon_deg + lon_min/60

            try:
                evdp=float(line[32:36])/100
            except:
                evdp = 0

            e_secs = float(_seconds)
            e_time = UTCDateTime(yr,mo,day,hr,minute,0)+e_secs

            str_time = e_time.strftime('%Y%m%d%H%M%S%f')
            str_time = str_time[:16]
            phs_dict[str_time] = {}
            phs_dict[str_time]["eve_loc"] = [evlo,evla,evdp]
            phs_dict[str_time]["phase"] = []
#            try:
#                evid = int(line[136:146])
#                phs_dict[str_time]["evid"] = evid
#            except:
#                print(f"Warning: no evid reads in for {event_count}")
#                pass

        elif re.match("[A-Z]+",f_para) and f_para != "  ": # phase line
            net = line[5:7]
            sta = re.split(" +",line[0:5])[0]
            year = int(line[17:21])
            month = int(line[21:23])
            day = int(line[23:25])
            hour = int(line[25:27])
            minute = int(line[27:29])
            if line[14]==" ":
                #if sec or msec is 0, it will be "  " in out.arc file
                _sec = line[41:44]; _sec_m = line[44:46]
                if _sec == "   ":
                    _sec = "000"
                if _sec_m == "  ":
                    _sec_m = "00"
                p_type="S"
                phs_time = UTCDateTime(year,month,day,hour,minute,0)+\
                                   (int(float(_sec))+int(_sec_m)*0.01)
                phs_dict[str_time]["phase"].append([net,sta,p_type,phs_time-e_time])
            
            else:
                p_type="P"
                #if sec or msec is 0, it will be "  " in out.arc file
                _sec = line[29:32]; _sec_m = line[32:34]
                if _sec == "   ":
                    _sec = "000"
                if _sec_m == "  ":
                    _sec_m = "00"
                phs_time = UTCDateTime(year,month,day,hour,minute,0)+\
                                   (int(float(_sec))+int(_sec_m)*0.01)
                phs_dict[str_time]["phase"].append([net,sta,p_type,phs_time-e_time])
        elif f_para=="  ":
            evid = int(line[66:72])
            phs_dict[str_time]["evid"]=evid
            
    out_name = y2000_file+".pkl"
    out_file = open(out_name,'wb')
    pickle.dump(phs_dict,out_file)
    out_file.close()
 
    return phs_dict


def phs_subset(phs_file,evid_list=[],loc_filter=[]):
    """
    subset the *.phs file by event id list or by region location
    len(evid_list) == 0 means no evid filter applied
    len(loc_filter) == 0 means no location filter applied

    Parameters:
        loc_filter in format [lon_min,lon_max,lat_min,lat_max]
    """
    evid_filt = False
    loc_filt = False
    
    if len(evid_list) > 0:
        evid_filt = True

    if len(loc_filter) > 0:
        loc_filt = True
        if len(loc_filter)!=4:
            raise Exception(f"Values qty in loc_filter is {len(loc_filter)},should be 4.")
        lon_min,lon_max,lat_min,lat_max = loc_filter

    cont = [] 
    with open(phs_file,'r') as f:
        for line in f:
            line = line.rstrip()
            cont.append(line)
    output = []
    tmp = []
    for line in cont:
        tmp.append(line)
        if re.match("\d+",line[:4]):
            lat = int(line[16:18])+int(line[19:23])*0.01/60
            if line[18] == "S":
                lat = -lat
            lon = int(line[23:26])+int(line[27:31])*0.01/60
            if line[26]=="W":
                lon = -lon
        if line[:4]== "    ": # last line of one event
            record_status = True
            evid = int(line[66:72])
            if evid_filt == True:
                if evid not in evid_list:
                    record_status = False
            if loc_filt == True:
                if lat<lat_min or lat>lat_max or lon<lon_min or lon>lon_max:
                    record_status = False

            if record_status == True:
                for line in tmp:
                    output.append(line)
            tmp = []               # reset to empty
                
    with open(phs_file+".sel",'w') as f:
        for line in output:
            f.write(line+"\n")
    f.close()


def load_sum_evid(sum_file):
    """
    *.sum file is the catalog summary file after Hyperinverse.
    This function returns a dict:
        -key is event id
        -value is an array with below component:
            --Str format event time "yyyymmddhhmmss**", also the event folder.
            --event longitude
            --event latitude
            --event depth
            --event magnitude
            --event travel time residual
    """
    sum_dict = {}
    with open(sum_file,'r') as f:
        for line in f:
            eve_id=int(line[136:146])
            eve_folder = line[0:16]
            evla = int(line[16:18])+0.01*int(line[19:23])/60
            evlo = int(line[23:26])+0.01*int(line[27:32])/60
            evdp = int(line[31:36])*0.01
            e_mag = int(line[123:126])*0.01
            e_res = int(line[48:52])*0.01
            sum_dict[eve_id] = [eve_folder,evlo,evla,evdp,e_mag,e_res]
    return sum_dict

def load_sum_evstr(sum_file):
    """
    *.sum file is the catalog summary file after Hyperinverse.
    This function returns a dict:
        -key is event time in "yyyymmddhhmmss**" format, same with event folder
        -value is an array with below component:
            --event id
            --event longitude
            --event latitude
            --event depth
            --event magnitude
            --event travel time residual
    """

    sum_dict = {}
    with open(sum_file,'r') as f:
        for line in f:
            eve_id=int(line[136:146])
            eve_folder = line[0:16]
            evla = int(line[16:18])+0.01*int(line[19:23])/60
            evlo = int(line[23:26])+0.01*int(line[27:32])/60
            evdp = int(line[31:36])*0.01
            e_mag = int(line[123:126])*0.01
            e_res = int(line[48:52])*0.01
            sum_dict[eve_folder] = [eve_id,evlo,evla,evdp,e_mag,e_res]
    f.close()
    return sum_dict

def arc_filt(arc_file="Y2000.phs",min_obs=8):
    cont = []
    with open(arc_file,"r") as f:
        cont = f.readlines()
    f.close()
    filt_cont = []
    tmp_cont = []

    for line in cont:
        if line[:4] != "    ":
            tmp_cont.append(line)
        else:
            tmp_cont.append(line)
            if len(tmp_cont) >= min_obs+1+1: # 1 event line, one evid line
                for line in tmp_cont:
                    filt_cont.append(line)
            tmp_cont = []

    with open(arc_file+".filt",'w') as f:
        for line in filt_cont:
            f.write(line)
    f.close()

class Hypoinv():
    def __init__(self,sum_file="out.sum"):
        self.dict_evid = load_sum_evid(sum_file)
        self.dict_evstr = load_sum_evstr(sum_file)
        self.get_locs()

    def get_locs(self):
        self.locs = []
        tmp = 0
        for key in self.dict_evid.keys():
            lon = self.dict_evid[key][1]
            lat = self.dict_evid[key][2]
            dep = self.dict_evid[key][3]
            mag = self.dict_evid[key][4]
            res = self.dict_evid[key][5]
            tmp = tmp+res
            self.locs.append([lon,lat,dep,mag])
        self.locs = np.array(self.locs)
        self.avg_res = tmp/self.locs.shape[0]
        
    def crop(self,lonmin,lonmax,latmin,latmax):
        pop_list = []
        for key in self.dict.keys():
            lon,lat,_,_,_ = self.dict[key]
            if lon<lonmin or lon>lonmax or lat<latmin or lat>latmax:
                pop_list.append(key)
                
        for key in pop_list:
            self.dict.pop(key)
    
    def hplot(self,
              xlim=[],
              ylim=[],
              markersize=6,
              size_ratio=1,
              imp_mag=3,
              add_cross=False,
              alonlat=[104,29],
              blonlat=[105,30],
              cross_width=0.1):

        plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    (self.locs[:,3]+2)*size_ratio,
                    edgecolors = "k",
                    facecolors='none',
                    marker='o',
                    alpha=1)

        kk = np.where(self.locs[:,3]>=imp_mag)
        if len(kk)>0:
            imp = plt.scatter(self.locs[kk,0],
                        self.locs[kk,1],
                        (self.locs[kk,3]+2)*size_ratio*5,
                        edgecolors ='r',
                        facecolors='none',
                        marker='*',
                        alpha=1)
            plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        if add_cross == True: # draw cross-section plot
            a1lon,a1lat,b1lon,b1lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=cross_width,
                                                   direction="right")
            a2lon,a2lat,b2lon,b2lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=cross_width,
                                                   direction="left")
            plt.plot([a1lon,b1lon,b2lon,a2lon,a1lon],
                     [a1lat,b1lat,b2lat,a2lat,a1lat],
                     linestyle='--',
                     c='darkred')
            plt.plot([alonlat[0],blonlat[0]],[alonlat[1],blonlat[1]],c='darkred')
        if len(xlim) != 0:
            plt.xlim(xlim)
        if len(ylim) != 0: 
            plt.ylim(ylim)
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        plt.show()
    
    def vplot(self,alonlat=[],blonlat=[],width=0.1,depmin=0,depmax=10,size_ratio=1,imp_mag=3):
        """
        Description
        """
        length_m,_,_ = gps2dist_azimuth(alonlat[1],alonlat[0],blonlat[1],blonlat[0])
        length_km = length_m/1000
        alon = alonlat[0]; alat = alonlat[1]
        blon = blonlat[0]; blat = blonlat[1]
        results = in_rectangle(self.locs,alon,alat,blon,blat,width)
        jj = np.where(results[:,0]>0)
        plt.scatter(results[jj,1],
                    self.locs[jj,2],
                    marker='o',
                    edgecolors = "k",
                    facecolors='none',
                    s=(self.locs[jj,3]+2)*size_ratio*5)
        tmplocs = self.locs[jj]
        tmpresults = results[jj]
        kk = np.where(tmplocs[:,3]>=imp_mag)

        if len(kk)>0:                 
            imp = plt.scatter(tmpresults[kk,1],
                        tmplocs[kk,2],
                        (tmplocs[kk,3]+2)*size_ratio*10,
                        edgecolors ='red',
                        facecolors='red',
                        marker='*',
                        alpha=1)
            plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        
        plt.ylim([depmax,depmin])
        plt.xlim([0,length_km])
        plt.xlabel("length (km)")
        plt.ylabel("depth (km)")
        plt.show()

    def depth_hist(self,mag_threshold=-9,depthmin=0,depthmax=10,gap=0.5):
        bins=np.arange(depthmin,depthmax,gap)
        fig,ax = plt.subplots(1,1,figsize=(6,8))
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Depth (km)",fontsize=16)
        ax.set_xlabel("Event Qty",fontsize=16)
        kk = np.where(self.locs[:,3]>=mag_threshold)
        hist,bins = np.histogram(self.locs[:,2],bins=bins)
        ax.barh(bins[:-1]+gap/2,hist,height=gap,color='gray',edgecolor='k')
        ax.set_ylim([depthmax,depthmin])
        plt.show()
        
    def __repr__(self):
        self.get_locs()
        _qty = f"Hypoinverse catlog with {len(self.locs)} events\n"
        _mag = f"Magnitue range is: {format(np.min(self.locs[:,3]),'4.1f')} to {format(np.max(self.locs[:,3]),'4.1f')}\n"
        _lon = f"Longitude range is: {format(np.min(self.locs[:,0]),'8.3f')} to {format(np.max(self.locs[:,0]),'8.3f')}\n"
        _lat = f"Latitude range is: {format(np.min(self.locs[:,1]),'7.3f')} to {format(np.max(self.locs[:,1]),'7.3f')}\n"
        _dep = f"Depth range is: {format(np.min(self.locs[:,2]),'4.1f')} to {format(np.max(self.locs[:,2]),'4.1f')}\n"
        return _qty+_mag+_lon+_lat+_dep
    
    def __getitem__(self,key):
        if isinstance(key,str):
            return self.dict_evstr[key]
        elif isinstance(key,int):
            return self.dict_evid[key]




# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
#     Author: Jinping ZI
#
# Revision History
#     2021-01-24 Initiate coding
#-----------------------------------------------------------------------------

import os
import numpy as np
from math import radians,cos,acos,sin,asin,sqrt,ceil,pi
import obspy
from obspy import Stream
import glob
import re
from obspy import UTCDateTime
import pandas as pd
from obspy.geodetics import gps2dist_azimuth
import logging
import pickle
from numba import cuda
from distutils.sysconfig import get_python_lib
import matplotlib.pyplot as plt

def add_path():
    print(">>> Add path to python library ...")
    pwd = os.getcwd()
    lib_path = get_python_lib()
    path_file = os.path.join(lib_path,'added.pth')
    with open(path_file,'a') as f:
        f.write(pwd)

    print("Done!")

cuda.jit()
def _matmul_gpu(A,B,C):
    row,col = cuda.grid(2)
    if row < C.shape[0] and col < C.shape[1]:
        tmp = 0.
        for k in range(A.shape[1]):
            tmp += A[row,k] * B[k,col]
        C[row,col] = tmp

def matmul_gpu(A,B,C,TPX=16,TPY=16):
    if len(cuda.gpus) == 0:
        raise Exception("Error, no gpu available")
    A_global_mem = cuda.to_device(A)
    B_global_mem = cuda.to_device(B)
    C_global_mem = cuda.device_array(A.shape[0],B.shape[1])
    threads_per_block = (TPA,TPB)
    blocks_per_grid_x = int(math.ceil(A.shape[0]/threads_per_block[0]))
    blocks_per_grid_y = int(math.ceil(B.shape[1]/threads_per_block[1]))
    blocks_per_grid = (blocks_per_grid_x,blocks_per_grid_y)

    _matmul_gpu[blocks_per_grid,blocks_per_block](A_global_mem,
                                                  B_global_mem,
                                                  C_global_mem)
    cuda.syschronize()
    C_global_gpu = C_global_mem.copy_to_host()


def init_logger(log_file,file_level=logging.DEBUG,stream_level=logging.INFO):
    '''
    Parameters:
        log_file: file to save the log information
        file_level: level for file writing
        stream_level: level for stream writing
    '''
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(log_file,mode='w')
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s-%(filename)s-%(levelname)s: %(message)s')
    fh.setFormatter(formatter)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.addHandler(fh)


def extract_set_info(pth,sta_file,depth=2):
    """
    Parameters:
        pth: path for the folder to work on
        depth: 1 or 2, 1 means pth/files, 2 means pth/folder/files
        
    Return:
        A setinfo dictionary containing keys:
        "s_times": sorted starttime list
        "e_times": sorted endtime list
        "netstas": network and station list in format <net+sta>
        "center" : mean longitude and latitude of stations,intended for tele-
                   event selection.
    """
    logger = logging.getLogger()
    logger.info("Extract_set_info program launched...")
    parent_path = os.path.split(pth)[0]
    if parent_path=="":
        parent_path="."
    if os.path.exists(os.path.join(parent_path,'setinfo.pkl')):
        f = open(os.path.join(parent_path,'setinfo.pkl'),'rb')
        setinfo = pickle.load(f)
        logger.info("Read in the existed setinfo.pkl and return")
        return setinfo

    setinfo = {}
    sta_dict = load_sta(sta_file)
    s_times = []
    e_times = []
    netstas = []
    stalons = []
    stalats = []
    logger.debug(f"file system depth is {depth}")
    availdays = []
    if depth == 2:                    # strcture: pth/folder/seis_file
        for item in os.listdir(pth):
            if os.path.isdir(os.path.join(pth,item)):
                logging.debug(f"Process dir {item}")
                juldays = []
                for file in os.listdir(os.path.join(pth,item)):
                    if file == "availdays.txt":
                        juldays = []
                        f = open(os.path.join(pth,item,file),'r')
                        for line in f:
                            line = line.rstrip()
                            julday = int(line)
                            juldays.append(julday)
                    try:
                        st = obspy.read(os.path.join(pth,item,file),headonly=True)
                        for tr in st:
                            s_julday = tr.stats.starttime.julday
                            e_julday = tr.stats.endtime.julday
                            if s_julday not in juldays:
                                juldays.append(s_julday)
                            if e_julday not in juldays:
                                juldays.append(e_julday)
                            net = tr.stats.network
                            sta = tr.stats.station
                            netsta = net+sta
                            if netsta not in netstas:
                                netstas.append(netsta)
                                stalons.append(sta_dict[net][sta][0])
                                stalats.append(sta_dict[net][sta][1])      
                            if tr.stats.starttime not in s_times:
                                s_times.append(tr.stats.starttime)
                            if tr.stats.endtime not in e_times:
                                e_times.append(tr.stats.endtime)
                    except:           # not seis file
                        pass
            else:
                continue
            if len(juldays) > 0:
                availdays.append(sorted(juldays))
    if len(availdays)>0 and len(availdays)!= len(netstas):
        raise Exception("len(availdays) not equal len(netstas)")
    setinfo["availdays"] = availdays
    setinfo["s_times"] = sorted(s_times)
    setinfo["e_times"] = sorted(e_times)
    setinfo["stalons"] = stalons
    setinfo["stalats"] = stalats
    setinfo["netstas"] = netstas
    setinfo["center"] = [np.mean(stalons),np.mean(stalats)]
    f = open(os.path.join(parent_path,"setinfo.pkl"),'wb')
    pickle.dump(setinfo,f)
    f.close()
    logging.info("extract set info programme done")
    return setinfo

def load_sta(sta_file):
    sta_dict={}
    with open(sta_file,'r') as f:
        for line in f:
            line = line.rstrip()
            net,sta,_lon,_lat,_ele,label=re.split(" +",line)
            if net not in sta_dict:
                sta_dict[net]={}
            if sta not in sta_dict[net]:
                sta_dict[net][sta] = [float(_lon),float(_lat),float(_ele),label]
    return sta_dict

def draw_vel(ax,dep_list,vel_list,color='k',linestyle='-'):
    """
    Draw velocity line on the ax based on the depth list and velocity list
    """
    points_list = []
    points_list.append([dep_list[0],vel_list[0]])
    for i in range(1,len(dep_list)):
        points_list.append([dep_list[i],vel_list[i-1]])
        points_list.append([dep_list[i],vel_list[i]])
        
    points_list = np.array(points_list)
    line, = ax.plot(points_list[:,1],points_list[:,0],color=color,linestyle=linestyle)
    return line


def read_sac_ref_time(tr):
    """
    Read and return reference time of a sac file in obspy.UTCDateTime format.

    Parameter
    --------
    tr: Trace object of obspy
    """

    nzyear = tr.stats.sac.nzyear
    nzjday = tr.stats.sac.nzjday
    nzhour = tr.stats.sac.nzhour
    nzmin = tr.stats.sac.nzmin
    nzsec = tr.stats.sac.nzsec
    nzmsec = tr.stats.sac.nzmsec*0.001
    year,month,day = month_day(nzyear,nzjday)
    sac_ref_time = UTCDateTime(year,month,day,nzhour,nzmin,nzsec)+nzmsec
    return sac_ref_time

def get_st(net,sta,starttime,endtime,f_folder,pad=False,fill_value=None):
    """
    Read and return waveform between starttime and endtime by specified
    net and station in designated folder. It will merge waveform if include
    more than one file.
    
    The return is a obspy Stream object
    """
    inc_list=[]
    for file in os.listdir(f_folder):
        file_path = os.path.join(f_folder,file)
        try:
            st = obspy.read(file_path,headonly=True)
        except:
            continue
        t1,t2 = st[0].stats.starttime,st[0].stats.endtime
        if t2 < starttime or t1 > endtime \
            or st[0].stats.network != net\
            or st[0].stats.station != sta:
            continue
        else:
            inc_list.append(file_path)
    #Read in data
    st = Stream()
    for path in inc_list:
        st += obspy.read(path)
    if len(st) == 0:
        pass
    else:
        st.trim(starttime,endtime,pad=pad,fill_value=fill_value)
    return st

def get_st_SC(net,sta,starttime,endtime,f_folder,pad=False,fill_value=None):
    """
    A modified get_st function for Agency data which stores data by UTM-8 and days.
    """
    ymd_start = starttime.strftime("%Y%m%d")
    ymd_end = endtime.strftime("%Y%m%d")
    st = Stream()
    if ymd_start == ymd_end:
        try:
            f_folder = os.path.join(f_folder,ymd_start[:4],ymd_start[4:6],ymd_start[6:])
            st += obspy.read(os.path.join(f_folder,f"*{sta}*"))
        except:
            pass
    else:
        try:
            f_folder = os.path.join(f_folder,ymd_start[:4],ymd_start[4:6],ymd_start[6:])
            st += obspy.read(os.path.join(f_folder,f"*{sta}*"))
            f_folder = os.path.join(f_folder,ymd_end[:4],ymd_end[4:6],ymd_end[6:])
            st += obspy.read(os.path.join(f_folder,f"*{sta}*"))
        except:
            pass
    if len(st) == 0:
        pass
    else:
        if len(st)>3:
            st = st.merge()
        if st[0].stats.starttime-8*60*60>=endtime or \
           st[0].stats.endtime-8*60*60<=starttime:
            st = Stream()
        else:
            st.trim(starttime-8*60*60,endtime-8*60*60,pad=pad,fill_value=fill_value)
    return st

def julday(year,month,day):
    ref_time=UTCDateTime(year,1,1)
    tar_time=UTCDateTime(year,month,day)
    julday=(tar_time-ref_time)/(24*60*60)+1
    return int(julday)

def month_day(year,julday):
    """
    Transfer from julday to month and day.
    Return year,month,day
    """
    #check if year is leap year
    leap=False
    if year%100==0:
        if year%400==0:
            leap=True
    else:
        if year%4==0:
            leap=True
    normal_list=[0,31,59,90,120,151,181,212,243,273,304,334,365]
    leap_list=[0,31,60,91,121,152,182,213,244,274,305,335,366]
    if leap:
        i=0
        while leap_list[i]<julday:
            i=i+1
        month=i
        day=julday-leap_list[i-1]
        return year,month,day
    else:
        i=0
        while normal_list[i]<julday:
            i=i+1
        month=i
        day=julday-normal_list[i-1]
        return year,month,day

def find_nearest(array,value):
    """
    find the nearest value. The return is index and diff
    """
    if type(array) != np.ndarray:
        array=np.array(array)
    idx=np.abs(array-value).argmin()
    diff=array[idx]-value
    return idx,diff

def time_interval_ticks(b_time,e_time,interval=1,unit='month',cut_to_round=True):
    """
    Generate month list from two UTCDateTime objects
    
    Parameters:
    |   b_time: begin time in obspy UTCDateTime format
    |   e_time: end time in obspy UTCDateTime format
    |     unit: interval unit, could be 'year','month','day','hour','minute',or 'second'
  cut_to_round: round off to the unit provided.default True.
                E.g. unit='month' will round the  base_time to be exactly UTCDateTime(year,month,1)
                else, base_time = b_time
    Return
    ------------------------------------------------
    base_time: UTCDateTime of the start of the first month
    tick_secs: array stores the tick points in seconds
    """
    assert e_time > b_time           # end time larger than begin time
    
    b_yr = b_time.year
    b_mo = b_time.month
    b_day = b_time.day
    b_hour= b_time.hour
    b_min = b_time.minute
    b_sec = b_time.second
    
    if cut_to_round == True:
        if unit == 'year':
            base_time = UTCDateTime(b_yr,1,1)
        if unit == 'month':
            base_time = UTCDateTime(b_yr,b_mo,1)
        elif unit == 'day':
            base_time = UTCDateTime(b_yr,b_mo,b_day)
        elif unit == 'hour':
            base_time = UTCDateTime(b_yr,b_mo,b_day,b_hour)
        elif unit == 'minute':
            base_time = UTCDateTime(b_yr,b_mo,b_day,b_hour,b_min)
        elif unit == 'second':
            base_time = UTCDateTime(b_yr,b_mo,b_day,b_hour,b_min,b_sec)
    else:
        base_time = b_time
    
    tick_secs = []
    loop_time = base_time
    
    if unit == 'year':
        while loop_time < e_time:
            tick_secs.append(loop_time - base_time)
            yr = loop_time.year
            mo = loop_time.month
            dy = loop_time.day
            hr = loop_time.hour
            minu = loop_time.minute
            sec = loop_time.second
            msec = loop_time.microsecond
            yr += interval
            loop_time = UTCDateTime(yr,mo,dy,hr,minu,sec,msec)
    elif unit == 'month':
        while loop_time < e_time:
            tick_secs.append(loop_time - base_time)
            yr = loop_time.year
            mo = loop_time.month
            dy = loop_time.day
            hr = loop_time.hour
            minu = loop_time.minute
            sec = loop_time.second
            msec = loop_time.microsecond
            mo += interval
            if mo > 12:
                yr += 1
                mo -= 12
            loop_time = UTCDateTime(yr,mo,dy,hr,minu,sec,msec)
    elif unit in ['day','hour','minute','second']:
        if unit == 'day':
            interval_seconds = interval*24*60*60
        elif unit == 'hour':
            interval_seconds = interval*60*60
        elif unit == 'minute':
            interval_seconds = interval*60
        elif unit == 'second':
            interval_seconds = interval

        while loop_time < e_time:
            tick_secs.append(loop_time - base_time)
            loop_time += interval_seconds
    else:
        raise Exception("'unit' not in ['year','month','day','hour','minute','second']")
        
    tick_secs.append(loop_time-base_time)

    return base_time, np.array(tick_secs)

def readfile(texfile):
    cont = []
    with open(texfile,'r') as f:
        for line in f:
            cont.append(line.rstrip())
    return cont

def writefile(cont,texfile):
    with open(texfile,'w') as f:
        for line in cont:
            f.write(line+"\n")

def bdy2pts(xmin,xmax,ymin,ymax):
    """
    Convert boundary to points for plot
    """
    points = []
    points.append([xmin,ymin])
    points.append([xmin,ymax])
    points.append([xmax,ymax])
    points.append([xmax,ymin])
    points.append([xmin,ymin])
    return np.array(points)
def matrix_show(*args,**kwargs):
    """
    Show matrix values in grids shape
    Parameters:cmap="cool",gridsize=0.6,fmt='.2f',label_data=True
    """
    ws = []
    H = 0
    str_count = 0
    ndarr_count = 0
    new_args = []
    for arg in args:
        if isinstance(arg,str):
            new_args.append(arg)
            continue
        if isinstance(arg,list):
            arg = np.array(arg)
        if len(arg.shape)>2:
            raise Exception("Only accept 2D array")
        if len(arg.shape) == 1:
            n = arg.shape[0]
            tmp = np.zeros((n,1))
            tmp[:,0] = arg.ravel()
            arg = tmp
        h,w = arg.shape
        if h>H:
            H=h
        ws.append(w)
        new_args.append(arg)
        ndarr_count += 1
    W = np.sum(ws)+len(ws)    # text+matrix+text+...+matrix+text
    if W<0:
        raise Exception("No matrix provided!")
        
    fmt = '.2f'
    grid_size = 0.6
    cmap = 'cool'
    label_data = True
    for arg in kwargs:
        if arg == "fmt":
            fmt = kwargs[arg]
        if arg == 'grid_size':
            grid_size = kwargs[arg]
        if arg == 'cmap':
            cmap = kwargs[arg]
        if arg == 'label_data':
            label_data = kwargs[arg]
    fig = plt.figure(figsize=(W*grid_size,H*grid_size))
    gs = fig.add_gridspec(nrows=H,ncols=W)
    
    wloop = 0
    matrix_id = 0
    for arg in new_args:
        if isinstance(arg,str):
            ax = fig.add_subplot(gs[0:H,wloop-1:wloop])
            ax.axis("off")
            ax.set_xlim(0,1)
            ax.set_ylim(0,H)
            ax.text(0.5,H/2,arg,horizontalalignment='center',verticalalignment='center')
        if isinstance(arg,np.ndarray):
            h,w = arg.shape
            hlow = int(np.round((H-h+0.01)/2))        # Find the height grid range
            hhigh = hlow+h
            wlow = wloop
            whigh = wlow+w
#            print("H: ",H,hlow,hhigh,"; W ",W,wlow,whigh)
            ax = fig.add_subplot(gs[hlow:hhigh,wlow:whigh])
            
            plt.pcolormesh(arg,cmap=cmap)
            for i in range(1,w):
                plt.axvline(i,color='k',linewidth=0.5)
            for j in range(1,h):
                plt.axhline(j,color='k',linewidth=0.5)
            if label_data:
                for i in range(h):
                    for j in range(w):
                        plt.text(j+0.5,i+0.5,format(arg[i,j],fmt),
                                 horizontalalignment='center',
                                 verticalalignment='center')
            plt.xlim(0,w)
            plt.ylim([h,0])
            plt.xticks([])
            plt.yticks([])
            wloop+=w+1
            matrix_id+=1
    plt.show()

def layer2linear_vel(inp_depths,inp_vels,linear_nodes):
    """
    Convert from layered 1-D model to linear velocity model, designed to conver from 
    the HYPOINVERSE model to tomoDD input model. It balances the average slowness.
    Parameters:
    | inp_depths: input depths, 1-D array/list
    | inp_vels: input velocities, 1-D array/list
    | linear_nodes: the depth nodes of output velocities
    
    Return:
    | average vlocity list correponding to the input velocity nodes

    Example:
    >>> inp_depths = [ -1,0.5,1.,1.5,2.13,4.13,6.9,10.07,15.93,17.,18.,27.,31.89]
    >>> inp_vels = [3.67,4.7,5.38,5.47,5.5,5.61,5.73,6.19,6.23,6.31,6.4,6.45,6.5]
    >>> tomo_deps = [0,1.5,3.0,4.5,6.0,7.5,9.0,10.5,12,18,24]
    >>> layer2linear_vel(inp_depths,inp_vels,tomo_deps)
    """
    
    #----------- quality control --------------------------------
    if linear_nodes[-1]>inp_depths[-1]:
        raise Exception("linear_nodes[-1] should less than inp_depths[-1]")
    if inp_depths.shape[0] != inp_vels.shape[0]:
        raise Eception("Different length of input depths quantity and velocity values")
    
    #----------- load in data ------------------------------------
    pxnodes = []
    pynodes = []
    for i in range(inp_depths.shape[0]):
        if i == 0:
            pxnodes.append(inp_vels[i])
            pynodes.append(inp_depths[i])
        else:
            pxnodes.append(inp_vels[i-1])
            pxnodes.append(inp_vels[i])
            pynodes.append(inp_depths[i])
            pynodes.append(inp_depths[i])

    #----------- calculate the mean velocity ----------------------
    avg_vels = []
    for i in range(len(linear_nodes)):
        dep = linear_nodes[i]
        #---------------get left and right range--------------
        if i == 0:                                   # first node
            dep_il = linear_nodes[i]
            dep_ir = (linear_nodes[i+1]+linear_nodes[i])/2
        elif i == len(linear_nodes) - 1:             # last node
            dep_il = (linear_nodes[i-1]+linear_nodes[i])/2
            dep_ir = linear_nodes[i]+(linear_nodes[i]-linear_nodes[i-1])/2
        else:
            dep_il = (linear_nodes[i-1]+linear_nodes[i])/2
            dep_ir = (linear_nodes[i+1]+linear_nodes[i])/2

        slw_list = []                                # slow list
        embrace_first = False
        for j in range(len(pynodes)):
            pynode = pynodes[j]
            velocity = pxnodes[j]                    # velocity

            if pynode >=dep_il: # consider velocity node within the range
                if embrace_first == False:
                    slw_list.append([dep_il,1/velocity]) # embrace the left node
                    embrace_first = True

                end_velocity = velocity
                if pynode>dep_ir:
                    break
                slw_list.append([pynode,1/velocity])
        slw_list.append([dep_ir,1/end_velocity])      # embrace the right node

        # calculate average slowness
        trav = 0                                      # travel time        
        for k in range(len(slw_list)-1):
            dep_a = slw_list[k][0]
            slw_a = slw_list[k][1]
            dep_b = slw_list[k+1][0]
            slw_b = slw_list[k+1][1]
            if (dep_b - dep_a)==0 or slw_a==slw_b:
                trav += (dep_b -dep_a)*slw_a
            else:
                raise Exception("(dep_b - dep_a)==0 or vel_a==vel_b")
        avg_vel = (dep_ir-dep_il)/trav
        avg_vels.append(avg_vel)
        
    return avg_vels



import re
from obspy.geodetics import gps2dist_azimuth
import json
import copy
import numpy as np
import matplotlib.pyplot as plt

def sta_dist_pairs(netstas,sta_dict,group_qty=6):
    """
    For station list provided, form unique pairs of stations by cloest
    distance.

    Parameters:
        netstas: station list in format network(2 char)+station(maximum 5 
                 character
        stadict: loaded station file containing station lon and lat information
      group_qty: control the quantity of cloest stations to form pairs.
    """
    pair_list = []
    for netsta1 in netstas:     # loop for each sta
        net1 = netsta1[:2]
        sta1 = netsta1[2:]
        lon1 = sta_dict[net1][sta1][0]
        lat1 = sta_dict[net1][sta1][1]
        dist_list = []
        for netsta2 in netstas: # calculate distance one by one
            net2 = netsta2[:2]
            sta2 = netsta2[2:]
            lon2 = sta_dict[net2][sta2][0]
            lat2 = sta_dict[net2][sta2][1]  
            dist,_,_ = gps2dist_azimuth(lat1,lon1,lat2,lon2)
            dist_list.append(dist)
        dist_list_cp = dist_list.copy()
        dist_list_cp.sort()
        for i in range(1,min([group_qty+1,len(netstas)])): # 0 idx is self
            dist = dist_list_cp[i]
            idx = dist_list.index(dist)
            tmp_netsta = netstas[idx]
            if [netsta1,tmp_netsta] not in pair_list and\
               [tmp_netsta,netsta1] not in pair_list:
                pair_list.append([netsta1,tmp_netsta])      
    return pair_list

def load_sta(sta_file):
    """
    Load in station information.

    Parameter:
        sta_file: text file in free format "net sta lon lat ele label\n"
        The label is intended for mark special stations purpose 
    Return:
        dictionary in structure sta_dict[net][sta]=[lon,lat,ele,label]
    """
    sta_dict={}
    with open(sta_file,'r') as f:
        for line in f:
            line = line.rstrip()
            net,sta,_lon,_lat,_ele,label=re.split("[ ,;]+",line)
            if net not in sta_dict:
                sta_dict[net]={}
            if sta not in sta_dict[net]:
                sta_dict[net][sta] = [float(_lon),float(_lat),int(float(_ele)),label]
    return sta_dict

def getNet(sta,staFile):
    staDict = load_sta(staFile)
    findStatus = False
    for net in staDict.keys():
        for key in staDict[net].keys():
            if key == sta:
                findStatus = True
                return net
    if findStatus == False:
        raise Exception("Sta not in station file.")
    return net

def to_inv_sta_file(sta_dict,out_file="sta.inv",ele_zero=True):
    f_inv = open(out_file,'w')
    for net in sta_dict.keys():
        for sta in sta_dict[net].keys():
            lon = sta_dict[net][sta][0]
            lon_marker="E"
            if lon < 0:
                lon_marker = "W"
                lon = -lon
            lat = sta_dict[net][sta][1]
            ele = sta_dict[net][sta][2]
            label = sta_dict[net][sta][3]
            net_sta = net+sta
            lon_i = int(lon)
            lon_f = lon-lon_i
            lat_i = int(lat)
            lat_f = lat-lat_i
            if ele_zero:
                ele = 0
            f_inv.write(format(sta,"<6s")+format(net,"<4s")+"SHZ  "+format(lat_i,">2d")+" "+\
                format(lat_f*60,">7.4f")+" "+format(lon_i,">3d")+" "+format(lon_f*60,">7.4f")+\
                lon_marker+format(ele,">4d")+"\n")
    f_inv.close()

def sta2inv(sta_file,out_file="sta.inv",ele_zero=True):
    """
    Convert station file into HYPOINVERSE format
    """
    sta_dict =load_sta(sta_file)
    to_inv_sta_file(sta_dict,out_file,ele_zero=True)  # Write into files

def to_dd_sta_file(sta_dict,out_file="sta.dd",ele_zero=True):
    f_dd = open(out_file,'w')
    for net in sta_dict.keys():
        for sta in sta_dict[net].keys():
            lon = sta_dict[net][sta][0]
            lat = sta_dict[net][sta][1]
            ele = sta_dict[net][sta][2]
            label = sta_dict[net][sta][3]
            net_sta = net+sta
            lon_i = int(lon)
            lon_f = lon-lon_i
            lat_i = int(lat)
            lat_f = lat-lat_i
            if ele_zero:
                ele = 0
            f_dd.write(format(net_sta,"<9s")+format(lat_i+lat_f,">9.6f")+format(lon_i+lon_f,">12.6f")+\
                   " "+format(ele,'>5d')+"\n")
    f_dd.close()

def sta2dd(sta_file,out_file="sta.dd",ele_zero=True):
    """
    Convert station file into hypoDD format
    """
    sta_dict = load_sta(sta_file)
    to_dd_sta_file(sta_dict,out_file,ele_zero=ele_zero)  # Write into files

def to_vel_sta_file(sta_dict,out_file="sta.vel",ele_zero=True):
    f_vel = open(out_file,'w')
    f_vel.write("(a5,f7.4,a1,1x,f8.4,a1,1x,i4,1x,i1,1x,i3,1x,f5.2,2x,f5.2,3x,i1)\n")
    sta_count = 1
    for net in sta_dict.keys():
        for sta in sta_dict[net].keys():
            lon = sta_dict[net][sta][0]
            lat = sta_dict[net][sta][1]
            ele = sta_dict[net][sta][2]
            label = sta_dict[net][sta][3]
            if ele_zero:
                ele = 0
            f_vel.write(f"{format(sta,'<5s')}{format(lat,'7.4f')}N {format(lon,'8.4f')}E {format(ele,'4d')} 1 "+\
                f"{format(sta_count,'3d')} {format(0,'5.2f')}  {format(0,'5.2f')}   1\n")
            sta_count += 1
    f_vel.write("  \n")   # signal of end of file for VELEST
    f_vel.close()

def sta2vel(sta_file,out_file="sta.vel",ele_zero=True):
    """
    Convert station file into VELEST format with 5 characters,
    which is applicable for the update VELEST program modified by Hardy ZI
    """
    sta_dict = load_sta(sta_file)
    to_vel_sta_file(sta_dict,out_file,ele_zero)

def toREAL(sta_dict,outFile):
    fsta = open(outFile,'w')
    for net in sta_dict.keys():
        for sta in sta_dict[net].keys():
            lon = sta_dict[net][sta][0]
            lat = sta_dict[net][sta][1]
            ele = sta_dict[net][sta][2]
            label = sta_dict[net][sta][3]
            fsta.write(f"{format(lon,'10.6f')} {format(lat,'10.6f')} {net} {format(sta,'5s')} BHZ {format(ele/1000,'5.3f')}\n")
    fsta.close()

def sta2REAL(sta_file,out_file):
    """
    Convert station file into VELEST format with 5 characters,
    which is applicable for the update VELEST program modified by Hardy ZI
    """
    sta_dict = load_sta(sta_file)
    toREAL(sta_dict,out_file)

def sta_sel(sta_file,c_lon,c_lat,nets=[],stas=[],radius=100):
    """
    select stations inside radius of a give lon and lat and output.
    output is a {sta_file}.sel file.
    Parameters:
        c_lon: longitude of the center. If c_lon<-180 or >180,radius filter will be passed
        c_lat: latitude of the center. If c_lat<-90 or >90, radisu filter will be passed
        nets: select nets if nets not empty
        stas: select stas if stas not empty 
    """
    select_net = False
    select_sta = False
    if len(nets)!=0:
        select_net = True
    if len(nets)!=0:
        select_sta = True

    out_file = sta_file+".sel"
    f1 = open(out_file,'w')
    with open(sta_file,'r') as f2:
        for line in f2:
            line = line.rstrip()
            net,sta,_lon,_lat,_ele,label = line.split()
            if len(nets)>0:   # select network
                if net not in nets:
                    continue
            if len(stas)>0:   # select station
                if sta not in stas:
                    continue
            lon = float(_lon)
            lat = float(_lat)
            if c_lon<=180 and c_lon>=-180 and c_lat>=-90 and c_lat<=90:
                dist,_,_ = gps2dist_azimuth(c_lat,c_lon,lat,lon)
                dist_km = dist/1000
                if dist_km > radius:
                    continue
            f1.write(line+"\n")
    f2.close()
    f1.close()

def sta2eqt(sta_file,out_file):
    """
    Convert station file into EQTransformer format
    """
    eqt_sta_dict = {}
    with open(sta_file,'r') as f:
        for line in f:
            line = line.rstrip()
            net,sta,_lon,_lat,_ele,_ = re.split(" +",line[:42])
            lon = float(_lon)
            lat = float(_lat)
            ele = float(_ele)
            eqt_sta_dict[sta]={}
            eqt_sta_dict[sta]["network"]=net
            eqt_sta_dict[sta]["channels"]=["BHN","BHE","BHZ"]
            eqt_sta_dict[sta]["coords"] = [lat,lon,ele]
    f.close()
    with open(out_file,'w') as dump_f:
        json.dump(eqt_sta_dict,dump_f)
    dump_f.close()


class Sta():
    def __init__(self,sta_file="/home/zijinping/Dropbox/resources/stations/sta_sum_202110.txt"):
        self.sta_file = sta_file
        self.dict = load_sta(sta_file)
        self.get_locs()
        
    def get_locs(self):
        self.locs = []
        for net in self.dict.keys():
            for sta in self.dict[net].keys():
                lon = self.dict[net][sta][0]
                lat = self.dict[net][sta][1]
                ele = self.dict[net][sta][2]
                marker = self.dict[net][sta][3]
                self.locs.append([lon,lat,ele])
        self.locs = np.array(self.locs)
        
    def plot(self,xlim=[],ylim=[],markersize=6,size=20,offsets=[0.1,0.1],fontsize=12):

        for net in self.dict.keys():
            for sta in self.dict[net].keys():
                lon = self.dict[net][sta][0]
                lat = self.dict[net][sta][1]
                if len(xlim) > 0:
                    if lon<xlim[0] or lon>xlim[1]:
                        continue
                if len(ylim) > 0:
                    if lat<ylim[0] or lat>ylim[1]:
                        continue
                plt.scatter(lon,
                    lat,
                    s=size,
                    edgecolors = "k",
                    facecolors='none',
                    marker='^',
                    alpha=1)
                plt.text(lon+offsets[0],
                    lat+offsets[1],
                    sta,
                    fontsize=fontsize)
                
        if len(xlim) != 0:
            plt.xlim(xlim)
        if len(ylim) != 0: 
            plt.ylim(ylim)
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        plt.show()
        
    def copy(self):
        return copy.deepcopy(self)
    
    def select(self,net=None):
        if net != None:
            dict_copy = self.dict.copy()
            self.dict = {}
            if isinstance(net,str):
                self.dict[net] = dict_copy[net]
            elif isinstance(net,list):
                for kk in net:
                    self.dict[kk] = dict_copy[kk]
        return self
    
    def subset(self,xlim=[],ylim=[]):
        dict_subset = {}
        for net in self.dict.keys():
            dict_subset[net] = {}
            for sta in self.dict[net].keys():
                lon = self.dict[net][sta][0]
                lat = self.dict[net][sta][1]
                if lon>=xlim[0] and lon<=xlim[1] and\
                   lat>=ylim[0] and lat<=ylim[1]:
                    dict_subset[net][sta] = self.dict[net][sta]
        self.dict = dict_subset
        
        return self
    
    def sta2inv(self):
        to_inv_sta_file(self.dict,ele_zero=True)
        
    def sta2dd(self):
        to_dd_sta_file(self.dict,ele_zero=True)
        
    def sta2vel(self):
        to_vel_sta_file(self.dict,ele_zero=True)
    
    def __repr__(self):
        self.get_locs()
        _qty = f"Including {len(self.locs)} stations\n"
        _lon = f"Longitude range is: {format(np.min(self.locs[:,0]),'8.3f')} to {format(np.max(self.locs[:,0]),'8.3f')}\n"
        _lat = f"Latitude range is: {format(np.min(self.locs[:,1]),'7.3f')} to {format(np.max(self.locs[:,1]),'7.3f')}\n"
        _dep = f"Elevation range is: {format(np.min(self.locs[:,2]),'4.1f')} to {format(np.max(self.locs[:,2]),'4.1f')}\n"
        return _qty+_lon+_lat+_dep
    
    def __getitem__(self,net):
            return self.dict[net]



import os
import copy
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
from obspy import UTCDateTime
from obspy.geodetics import gps2dist_azimuth
from seisloc.dd import loadDD
from seisloc.geometry import in_rectangle,loc_by_width
from math import floor,ceil
from seisloc.statistics import sum_count_Mo,write_sum_count_Mo
from seisloc.phase_convert import cata2fdsn
import logging


class Catalog():
    def __init__(self,locfile="hypoDD.reloc"):
        """
        The programme will read in hypoDD relocation file by default. If no hypoDD
        file provided (locfile=None), it will generate an empty catalog. 
        A user can set up a new catalog by providing a dict in the form:
            dict[evid] = [lon,lat,dep,mag,UTCDateTime]
        example:
        >>> cata = Catalog(locfile=None)
        >>> cata.dict = cata_dict  # cata_dict is a dictionary follows above format
        >>> cata.init()            # initiation of the class
        >>> print(cata)            # basic information will be printed
        """
        if locfile != None:
            if not os.path.exists(locfile):
                raise Exception(f"{locfile} not existed!")
            self.dict,_ = loadDD(locfile)
            print("successfully load catalog file: "+locfile)
            self.init()
        else:
            print("*** No hypoDD data provided, an empty Catalog created.")
            print("*** You can define self.dict[evid] = [lon,lat,dep,mag,UTCDateTime]}")
            print("*** Then run: .init() to initiate the catalog.")
            self.dict = {}

    def init(self):
        """
        Initiate the catalog
        """
        print(">>> Initiate the catalog ... ")
        self.init_keys()
        self.init_locs()
        self.init_relative_seconds()
        self.sort()
        print(">>> Initiation complted! ")

    def init_keys(self):
        """
        Build up event ids array
        """
        self.keys = list(self.dict.keys())
        self.keys = np.array(self.keys)
        self.keys = self.keys.astype(int)

    def init_locs(self):
        """
        Generate numpy array in format lon, lat, dep, mag
        """
        self.locs = []
        for key in self.keys:
            lon = self.dict[key][0]
            lat = self.dict[key][1]
            dep = self.dict[key][2]
            mag = self.dict[key][3]
            self.locs.append([lon,lat,dep,mag])
        self.locs = np.array(self.locs)

    def init_relative_seconds(self):
        """
        Numpy array to save relative seconds to the first event
        """
        self.first_key = self.keys[0]
        self.first_time = self.dict[self.first_key][4]
        self.relative_seconds = []
        for key in self.keys:
            etime = self.dict[key][4]
            self.relative_seconds.append(etime-self.first_time)
        self.relative_seconds = np.array(self.relative_seconds)

    def update_keys(self,idxs):
        """
        Update keys array with indexs
        """
        self.keys = self.keys[idxs]

    def update_dict(self):
        """
        Update dictionary with new set of keys
        """
        old_keys = list(self.dict.keys())
        for key in old_keys:
            if key not in self.keys:
                self.dict.pop(key)

    def update_locs(self,idxs):
        """
        Update location array with indexs
        """
        self.locs = self.locs[idxs]

    def update_relative_seconds(self,idxs):
        """
        Update relative times with indexs
        """
        self.relative_seconds = self.relative_seconds[idxs]
        
    def crop(self,lonmin,lonmax,latmin,latmax):
        """
        Trim the dataset with the lon-lat boundary conditions
        """
        idxs = np.where((self.locs[:,0]>=lonmin)&(self.locs[:,0]<=lonmax)&\
                        (self.locs[:,1]>=latmin)&(self.locs[:,1]<=latmax))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def magsel(self,mag_low,mag_top=10):
        """
        Select the dataset with the magnitude
        """
        idxs = np.where((self.locs[:,3]>=mag_low)&(self.locs[:,3]<=mag_top))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def trim(self,starttime,endtime):
        """
        Trim the dataset with time conditions
        """
        min_reftime = starttime - self.first_time
        max_reftime = endtime - self.first_time
        
        idxs = np.where((self.relative_seconds>=min_reftime)&\
                        (self.relative_seconds<=max_reftime))
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def sort(self,method="time"):
        idxs = self.relative_seconds.argsort()
        self.update_keys(idxs)
        self.update_dict()
        self.update_locs(idxs)
        self.update_relative_seconds(idxs)

    def hplot(self,
              xlim=[],
              ylim=[],
              figsize=None,
              edgecolor='grey',
              markersize=6,
              size_ratio=1,
              imp_mag=None,
              reftime = None,
              cmap = None,
              vmin=0,
              vmax=1,
              unit="day",
              add_section=False,
              alonlat=[104,29],
              blonlat=[105,30],
              section_width=0.05,
              crop=False,
              ax = None,
              plt_show=True):
        """
        Map view plot of earthquakes,earthquake denoted default by black circle
        Parameters:
        |         xlim: longitude limit, e.g. [104,105]
        |         ylim: latitude limit, e.g. [29,30]
        |      figsize: e.g. (5,5)
        |    edgecolor: earthquake marker(circle) edgecolor
        |      imp_mag: important magnitude. Magnitude larger than this level will be 
        |               highlighted
        |     reftime: reference time in UTCDateTime used to constrain colormap, if
        |               no colormap provided, seismicity will plotted by default
        |         cmap: colormap, check 'matplotlib' for more detail.
        |    vmin,vmax: the minimum and maximum value for colormap
        |         unit: "day","hour", or "second" for vmin and vmax
        |  add_section: if want to add one cross-section, set True
        |      alonlat: the [lon,lat] of the section start point 'a'
        |      blonlat: the [lon,lat] of the section end point 'b'
        |section_width: width of section in degree
        |         crop: if True, the dataset will cut dataset to leave only events
                        inside the cross-section
        """
        #---------------- initiation ----------------------------------
        if ax != None:     # get current axis
            plt.sca(ax)
        else:
            if figsize != None:
                plt.figure(figsize=figsize)
        if reftime == None:
            reftime = self.first_time + np.min(self.relative_seconds)
        if section_width <=0:
            raise Error("Width <= 0")
        # plot all events
        if add_section==True:
            alon = alonlat[0]; alat = alonlat[1]
            blon = blonlat[0]; blat = blonlat[1]
            results = in_rectangle(self.locs,alon,alat,blon,blat,section_width/2)
            jj = np.where(results[:,0]==1)
            if crop == True:
                self.update_keys(jj)
                self.update_relative_seconds(jj)
                self.update_locs(jj)
                self.update_dict()
                
        if cmap == None:
            plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    (self.locs[:,3]+2)*size_ratio,
                    edgecolors = edgecolor,
                    facecolors='none',
                    marker='o',
                    alpha=1)
        else:
            shift_seconds = reftime - self.first_time
            times_plot = self.relative_seconds-shift_seconds
            if unit=="day":
                times_plot = times_plot/(24*60*60)
            elif unit=="hour":
                times_plot = times_plot/(60*60)
            elif unit=="minute":
                times_plot = times_plot/60
            im = plt.scatter(self.locs[:,0],
                    self.locs[:,1],
                    c=times_plot,
                    s=(self.locs[:,3]+2)*size_ratio,
                    cmap = cmap,
                    vmin = vmin,
                    vmax = vmax,
                    marker='o',
                    alpha=1)
            cb = plt.colorbar(im)
            cb.set_label(unit)

        # plot large events
        if imp_mag != None:
            kk = np.where(self.locs[:,3]>=imp_mag)
            if len(kk)>0:                 
                imp = plt.scatter(self.locs[kk,0],
                        self.locs[kk,1],
                        (self.locs[kk,3]+2)*size_ratio*20,
                        edgecolors ='black',
                        facecolors='red',
                        marker='*',
                        alpha=1)
                plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        
        if add_section == True: # draw cross-section plot
            a1lon,a1lat,b1lon,b1lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=section_width/2,
                                                   direction="right")
            a2lon,a2lat,b2lon,b2lat = loc_by_width(alonlat[0],
                                                   alonlat[1],
                                                   blonlat[0],
                                                   blonlat[1],
                                                   width=section_width/2,
                                                   direction="left")
            plt.plot([a1lon,b1lon,b2lon,a2lon,a1lon],
                     [a1lat,b1lat,b2lat,a2lat,a1lat],
                     linestyle='--',
                     c='darkred')
            plt.plot([alonlat[0],blonlat[0]],[alonlat[1],blonlat[1]],c='darkred')
        # adjust plot parameters
        if len(xlim) != 0:
            plt.xlim(xlim)
        if len(ylim) != 0: 
            plt.ylim(ylim)
        plt.xlabel("Longitude")
        plt.ylabel("Latitude")
        plt.gca().set_aspect("equal")

        if plt_show == True:
            plt.show()
        
    def vplot(self,
              alonlat,
              blonlat,
              width=0.1,
              edgecolor='grey',
              depmin=0,
              depmax=10,
              size_ratio=1,
              imp_mag=None,
              reftime = UTCDateTime(2019,3,1),
              cmap=None,
              vmin=0,
              vmax=1,
              unit="day",
              aspect="auto",
              plt_show=True):
        """
        Description

        Parameters
        |   alonlat: the [lon,lat] of the section start point 'a'
        |   blonlat: the [lon,lat] of the section end point 'b'
        |     width: width of section in degree
        |    depmin: minimum depth in km, e.g. 0  km
        |    depmax: maximum depth in km, e.g. 10 km
        |   figsize: e.g. (5,5). Default None means auto set by matplotlib
        | edgecolor: earthquake marker(circle) edgecolor
        |   imp_mag: important magnitude. Magnitude larger than this level will be 
        |            highlighted
        |  reftime: reference time in UTCDateTime used to constrain colormap, if
        |            no colormap provided, seismicity will plotted by default
        |      cmap: colormap, check 'matplotlib' for more detail.
        | vmin,vmax: the minimum and maximum value for colormap
        |      unit: "day","hour", or "second" for vmin and vmax
        |    aspect: aspect ratio setting. Check for plt.gca().set_aspect for detail
        """
        length_m,_,_ = gps2dist_azimuth(alonlat[1],alonlat[0],blonlat[1],blonlat[0])
        length_km = length_m/1000
        alon = alonlat[0]; alat = alonlat[1]
        blon = blonlat[0]; blat = blonlat[1]
        results = in_rectangle(self.locs,alon,alat,blon,blat,width/2)
        jj = np.where(results[:,0]>0)
        self.vxy=np.zeros((jj[0].shape[-1],2))
        self.vkeys=np.zeros((jj[0].shape[-1],))
        self.vkeys = self.keys[jj].ravel()
        self.vxy[:,0] = results[jj,1]
        self.vxy[:,1] = self.locs[jj,2]
        if cmap==None:
            plt.scatter(results[jj,1],
                    self.locs[jj,2],
                    marker='o',
                    edgecolors = edgecolor,
                    facecolors='none',
                    s=(self.locs[jj,3]+2)*size_ratio*5)
        else:
            shift_seconds = reftime - self.first_time
            times_plot = self.relative_seconds[jj]-shift_seconds
            if unit=="day":
                times_plot = times_plot/(24*60*60)
            elif unit=="hour":
                times_plot = times_plot/(60*60)
            elif unit=="minute":
                times_plot = times_plot/60
            im = plt.scatter(results[jj,1],
                    self.locs[jj,2],
                    c=times_plot,
                    s=(self.locs[jj,3]+2)*size_ratio*5,
                    cmap = cmap,
                    vmin = vmin,
                    vmax = vmax,
                    marker='o',
                    alpha=1)
            cb = plt.colorbar(im)
            cb.set_label(unit)

        tmplocs = self.locs[jj]
        tmpresults = results[jj]
        if imp_mag != None:
            kk = np.where(tmplocs[:,3]>=imp_mag)
            if len(kk)>0:                 
                imp = plt.scatter(tmpresults[kk,1],
                        tmplocs[kk,2],
                        (tmplocs[kk,3]+2)*size_ratio*30,
                        edgecolors ='black',
                        facecolors='red',
                        marker='*',
                        alpha=1)
                plt.legend([imp],[f"M$\geq${format(imp_mag,'4.1f')}"])
        
        plt.ylim([depmax,depmin])
        plt.xlim([0,length_km])
        plt.xlabel("distance (km)")
        plt.ylabel("depth (km)")
        plt.gca().set_aspect(aspect)

        if plt_show==True:
            plt.show()
    
    def MTplot(self,
                xlim=[],
                ylim=[0,5],
                unit="day",
                reftime=None,
                cmap=None,
                vmin=0,
                vmax=1,
                figsize=(10,5),
                plt_show=True):
        """
        unit: 'day','hour' or 'second'
        """
        # ----------------- initiate -----------------------------------
        fig,ax = plt.subplots(1,figsize=figsize)
        if unit == "day":
            denominator = (24*60*60)
            plt.xlabel("Time (day)")
        elif unit == "hour":
            denominator = (60*60)
            plt.xlabel("Time (hour)")
        elif unit == "second":
            denominator = 1
            plt.xlabel("Time (second)")
        if reftime == None:
            reftime = self.first_time + np.min(self.relative_seconds)
        print("Reference time is: ", reftime)
        if len(xlim) == 0:
            tbtime = self.first_time + np.min(self.relative_seconds)
            tetime = self.first_time + np.max(self.relative_seconds)
        if len(xlim) == 1:
            tbtime = reftime+xlim[0]*denominator
            tetime = self.first_time + np.max(self.relative_seconds)
        if len(xlim) == 2:
            tbtime = reftime+xlim[0]*denominator
            tetime = reftime+xlim[1]*denominator

        #----------------- processing ---------------------------------
        for key in self.keys:
            etime = self.dict[key][4]
            if etime< tbtime or etime>tetime:
                continue
            emag = self.dict[key][3]
            diff_seconds = etime - reftime
            diff_x = diff_seconds/denominator
            if cmap == None:
                plt.plot([diff_x,diff_x],[ylim[0],emag],c='grey')
            else:
                plt.plot([diff_x,diff_x],[ylim[0],emag],color=cmap((diff_x-vmin)/(vmax-vmin)))
            plt.plot([diff_x],emag,'x',c='k')
        plt.ylim(ylim)
        if len(xlim)>0:
            plt.xlim(xlim)
        else:
            plt.xlim(left = (tbtime-reftime)/denominator-0.1,right = (tetime-reftime)/denominator+0.1)
        plt.ylabel("Magnitude")

        if plt_show == True:
            plt.show()

    def dep_dist_plot(self,refid=None,
                  refloc = [],
                  reftime=None,
                  xlim=[],
                  deplim=[100,-4],
                  distlim=[],
                  unit="day",
                  cmap=None,
                  vmin=0,
                  vmax=1,
                  figsize=(8,6),
                  plt_show=True):
        #------------- initiation --------------------
        fig,axs = plt.subplots(2,1,figsize=figsize)
        if unit == "day":
            denominator = (24*60*60)
            plt.xlabel("Time (day)")
        elif unit == "hour":
            denominator = (60*60)
            plt.xlabel("Time (hour)")
        elif unit == "second":
            denominator = 1
            plt.xlabel("Time (second)")
        if len(xlim)>0:
            axs[0].set_xlim(xlim)
            axs[1].set_xlim(xlim)
        axs[0].set_ylabel("Depth (km)")
        axs[1].set_ylabel("3D-dist (km)")
        axs[0].set_ylim(deplim)
        if len(distlim)>0:
            axs[1].set_ylim(distlim)
        axs[0].grid(axis="y")
        axs[1].grid(axis="y")
        if refid != None:
            reflon = self.dict[refid][0]
            reflat = self.dict[refid][1]
            refdep = self.dict[refid][2]
        if len(refloc)>0:
            reflon,reflat,refdep = refloc
        if reftime == None:
            reftime = self.first_time + np.min(self.relative_seconds)

        for evid in self.keys:
            etime =self.dict[evid][4]
            elon = self.dict[evid][0]
            elat = self.dict[evid][1]
            edep = self.dict[evid][2]
            emag = self.dict[evid][3]
            diff_x = (etime-reftime)/denominator
            dist,_,_ = gps2dist_azimuth(elat,elon,reflat,reflon)
            d3dist = np.sqrt((dist/1000)**2+(edep-refdep)**2)
            if cmap==None:
                axs[0].scatter(diff_x,edep,s=(emag+2)*5,marker='o',c='k')
                axs[1].scatter(diff_x,d3dist,s=(emag+2)*5,marker='o',c='k')
            else:
                axs[0].scatter(diff_x,edep,s=(emag+2)*5,marker='o',color=cmap((diff_x-vmin)/(vmax-vmin)))
                axs[1].scatter(diff_x,d3dist,s=(emag+2)*5,marker='o',color=cmap((diff_x-vmin)/(vmax-vmin)))
    
        plt.tight_layout()
        if plt_show == True:
            plt.show()
        
    def depth_hist(self,mag_threshold=-9,depthmin=0,depthmax=10,gap=0.5):
        bins=np.arange(depthmin,depthmax,gap)
        fig,ax = plt.subplots(1,1,figsize=(6,8))
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")
        ax.set_ylabel("Depth (km)",fontsize=16)
        ax.set_xlabel("Event Qty",fontsize=16)
        kk = np.where(self.locs[:,3]>=mag_threshold)
        hist,bins = np.histogram(self.locs[:,2],bins=bins)
        ax.barh(bins[:-1]+gap/2,hist,height=gap,color='gray',edgecolor='k')
        ax.set_ylim([depthmax,depthmin])
        return ax
        
    def day_hist(self,
                 reftime=UTCDateTime(2019,1,1,0,0,0),
                 xlim=[],
                 ylim=[],
                 color='b',
                 edgecolor='k',
                 plot_months=True,
                 figsize=None,
                 plt_show = True):
        """
        Plot events by day-quantity in a histogram plot.
        Parameters:
            -reftime: Reference time for plot
        """
        ref_list = []
        time_list = []
        for key in self.dict.keys():
            _,_,_,_,etime = self.dict[key]
            ref_list.append((etime-reftime)/(24*60*60))

        min_day=floor(min(ref_list))
        max_day=ceil(max(ref_list))
        bins = np.linspace(min_day,max_day,max_day-min_day+1)
        if figsize==None:
            figsize=(8,4)
        fig1 = plt.figure(1,figsize=figsize)
        ax1 = plt.subplot(1,1,1)
        ax1.hist(ref_list,bins,color=color,edgecolor=edgecolor)
        # The bottom x-axis is in days
        ax1.set_xlim([0,max_day])
        # The top x-axis marks year and month in YYYYMM
        tick_list_1 = [] # Store the position number
        tick_list_2 = [] # Store the tick text
        ref_year = reftime.year
        ref_month = reftime.month
        ref_day = reftime.day
        if ref_day == 1:
            tick_list_1.append(0)
            tick_list_2.append(str(ref_year)+str(ref_month).zfill(2))
        status = True # Start to loop month by month
        loop_time = UTCDateTime(ref_year,ref_month,1) # Initiate loop time
        step = 32 #32 > 31. Make sure each step pass to next month
        while status==True:
            loop_time = loop_time + step*24*60*60
            tmp_year = loop_time.year
            tmp_month = loop_time.month
            loop_time = UTCDateTime(tmp_year,tmp_month,1)
            diff_days = (loop_time - reftime)/(24*60*60)
            if diff_days > (max_day):
                status=False
            else:
                tick_list_1.append(diff_days)
                tick_list_2.append((str(tmp_month).zfill(2)))
        if plot_months:
            ax2 = ax1.twiny()
            ax2.set_xlim([0,max_day])
            ax2.plot(0,0,'k.')
            plt.xticks(tick_list_1,tick_list_2)
            ax2.set_xlabel("date")
        if xlim!=[]:
            ax1.set_xlim(xlim)
            if plot_months:
                ax2.set_xlim(xlim)
        if ylim!=[]:
            plt.ylim(ylim)
        ax1.set_xlabel("Time, days")
        ax1.set_ylabel("event quantity")

        if plt_show == True:
            plt_show()

    def diffusion_plot(self,refid=None,refloc=[],diff_cfs=[],unit="day",xlim=[],ylim=[],plt_show=True):
        '''
        Parameters:
        refid: reference event id
        refloc: [lon,lat], reference site longitude and latitude, if not provided, use refid
        diff_cfs: diffusion coefficient list, this will draw corresponding lines on the map
        '''
        #from seisloc.plot import diffusion_plot
        
        if refid==None and refloc==[]:
            raise Exception("refid or refloc should be proivded")
        if refloc==[]:
            refloc=[self.dict[refid][0],self.dict[refid][1]]
        dist_list = np.zeros((len(self.keys),1))
        day_list = np.zeros((len(self.keys),1))
        mag_list = np.zeros((len(self.keys),1))
        for i in range(len(self.keys)):
            dist,_,_ = gps2dist_azimuth(self.locs[i,1],\
                                self.locs[i,0],\
                                refloc[1],\
                                refloc[0])
            day_list[i,0] = (self.relative_seconds[i]-np.min(self.relative_seconds))/(24*60*60)
            mag_list[i,0] = self.locs[i,3]
            dist_list[i,0] = dist

        fig1 = plt.figure(1)
        ax1 = plt.subplot(1,1,1)
        if unit=="day":
            x_list = day_list
            plt.xlabel("Time (day)",fontsize=16)
        elif unit == "hour":
            x_list = day_list*24
            plt.xlabel("Time (hour)",fontsize=16)
        else:
            raise Exception("Unit error: 'day' or 'hour'")
        ax1.set_ylabel("Distance (m)",fontsize=16)
        ax1.scatter(x_list,dist_list,(mag_list+2)*3,c='k')
        ax1.set_xlim([0,np.max(x_list)])

        diff_lines = []
        if isinstance(diff_cfs,int) or isinstance(diff_cfs,float):
            diff_cfs=[diff_cfs]
        for diff_cf in diff_cfs:
            if unit=="day":
                x = np.linspace(0,np.max(x_list),int(np.max(x_list)*20)+1)
                y = np.sqrt(4*np.pi*diff_cf*x*24*60*60)
            elif unit=="hour":
                x = np.linspace(0,np.max(x_list),int(np.max(x_list)*20)+1)
                y = np.sqrt(4*np.pi*diff_cf*x*60*60)
            diff_line, = plt.plot(x,y)
            diff_lines.append(diff_line)

        plt.legend(diff_lines,diff_cfs,title="Diffusion Coefficient $m^2/s$")
        if len(xlim)>0:
            plt.xlim(xlim)
        if len(ylim)>0:
            plt.ylim(ylim)
        else:
            plt.ylim(bottom=0)
        if plt_show == True:
            plt_show()

    def animation(self,
                  incre_hour=2,
                  mb_time=None,
                  me_time=None,
                  xlim=[],
                  ylim=[],
                  geopara=None,
                  cmap=None,
                  vmin=None,
                  vmax=None):
        """
        Generate gif animation file
        increment: Time increased for each plot. Unit: hour
        """
        # Remove previous results
        try:
            shutil.rmtree("dd_animation")
        except:
            pass
        os.makedirs("dd_animation")
        if xlim == []:
            xlim = [np.min(self.locs[:,0]),np.max(self.locs[:,0])]
        if ylim == []:
            ylim = [np.min(self.locs[:,1]),np.max(self.locs[:,1])]    
        min_time = self.first_time+np.min(self.relative_seconds)
        max_time = self.first_time+np.max(self.relative_seconds)
        
        if mb_time == None:
            mb_time = min_time
        if me_time == None:
            me_time = max_time
        print("Movie start time is: ",mb_time)
        print("  Movie end time is: ",me_time)
        #---------- cmap -------------------------
        if vmin == None:
            vmin = 0
        if vmax == None:
            vmax = (max_time - min_time)/(24*60*60)

        inc_second = incre_hour*60*60 # Time increment
        loop_time = mb_time
        count = 1
        reftime = mb_time
        while loop_time <= me_time:
            fig = plt.figure(1,figsize=(8,8))
            ax1 = fig.add_subplot(111)
            ax1.set_xlim(xlim)
            ax1.set_ylim(ylim)
            ax1.set_xlabel("Lon(degree)",fontsize=18)
            ax1.set_ylabel("Lat(degree)",fontsize=18)
            ax1.set_title(f"{str(loop_time)[:19]}",fontsize=16)
            if geopara != None:
                # ----------------Molin faults--------------------------------
                ml_fault = np.array(geopara.dict['ml_fault'])
                ML_fault,=ax1.plot(ml_fault[:,0],ml_fault[:,1],'r-')
                # ----------------Zigong faults-------------------------------
                for key in geopara.dict['zg_faults']:
                    array = np.array(geopara.dict['zg_faults'][key])
                    ax1.plot(array[:,0],array[:,1],'k-',label='Faults')
                #----------------------Wells-----------------------------------
                wells = np.array(geopara.dict['wells'])
                well_lons=[]; well_lats=[]
                for well in geopara.dict['wells']:
                    well_lons.append(well[0])
                    well_lats.append(well[1])
                s_well,=ax1.plot(well_lons,well_lats,'s',c='#1f77b4',markerfacecolor='white',mew=2,markersize=12)
                #----------------------Stations--------------------------------
                sta_lons=[]; sta_lats=[]
                for sta in geopara.dict['sta_locs']:
                    if sta[3]=="SC":
                        sta_lons.append(sta[0]); sta_lats.append(sta[1])
                if len(sta_lons)>0:
                    s_sta=ax1.scatter(sta_lons,sta_lats,marker='^',c='cyan',s=120,edgecolor='k',label='Stations')
            #------------- Events -----------------------------------------
            eve_arr = []
            rela_days = []
            for i,second in enumerate(self.relative_seconds):
                e_time = self.first_time + second
                if e_time<(loop_time+inc_second/2) and e_time>mb_time:
                    eve_arr.append(self.locs[i,:4])
                    rela_days.append((e_time-reftime)/(24*60*60))
            eve_arr = np.array(eve_arr)
            rela_days = np.array(rela_days)

            if len(eve_arr)>0:
                if cmap != None:
                    s_eve=ax1.scatter(eve_arr[:,0],
                                          eve_arr[:,1],
                                          s=(eve_arr[:,3]+2)*5,
                                          c=rela_days,
                                          cmap=cmap,
                                          vmin=vmin,
                                          vmax=vmax,
                                          label="Events")
                else:
                    s_eve=ax1.scatter(eve_arr[:,0],
                                          eve_arr[:,1],
                                          s=(eve_arr[:,3]+2)*5,
                                          c='k',
                                          label="Events")
            if geopara != None:
                plt.legend([s_well,s_sta,s_eve],\
                       ["Platform",'Station',"Seismicity"],\
                       loc='upper right',
                       fontsize=16)  
            ##------------- save results --------------------------------------------
            plt.savefig(f"dd_animation/{str(count).zfill(3)}.png")
            loop_time = loop_time + inc_second
            count+=1
            plt.close()
        #-------------------- gif -----------------------------
        imgs = []
        for i in range(1,count):
            pic_name = f'dd_animation/{str(i).zfill(3)}.png'
            tmp = Image.open(pic_name)
            imgs.append(tmp)
        imgs[0].save("dd_animation.gif",save_all=True,append_images=imgs,duration=10)
    
    def intervals_plot(self,interval=1,method='month',
                       xlim=[],ylim=[],
                       columns=4,subplotsize=(3,3),
                       marker='o',ms=1,
                       wspace=None,hspace=None):
        
        from seisloc.plot import intervals_plot
        plt.close()

        axs = intervals_plot(xys=self.locs[:,:2],
                rela_secs=self.relative_seconds,
                reftime=self.first_time,
                interval=interval,method=method,
                xlim=xlim,ylim=ylim,
                columns=columns,subplotsize=subplotsize,
                marker=marker,ms=ms,
                wspace=wspace,hspace=hspace)
        
        return axs
   
    def depths_plot(self,
                deplim=[0,10],interval=1,
                xlim=[],ylim=[],
                columns=4,subplotsize=(3,3),
                marker='o',ms=1,color='k',
                zorder=0,
                wspace=None,hspace=None):
        from seisloc.plot import depths_plot
        axs = depths_plot(xyz=self.locs[:,:3],
                          deplim=deplim,interval=interval,
                          xlim=xlim,ylim=ylim,
                          columns=columns,subplotsize=subplotsize,
                          marker=marker,ms=ms,color=color,
                          zorder=zorder,
                          wspace=None,hspace=None)
        return axs
    def sum_count_Mo(self,starttime,endtime):
        self.dict_count_Mo = sum_count_Mo(self,starttime,endtime)
    def write_sum_count_Mo(self,outfile="sum_count_Mo.txt",mode='day'):
        write_sum_count_Mo(self.dict_count_Mo,outfile,mode)

    def write_info(self,info_file="lon_lat_dep_mag_relative_days.txt",reftime=None,disp=False):
        if reftime == None:
            reftime = self.first_time
        print("The reference time is: ",reftime)
        f = open(info_file,'w')
        for key in self.keys:
            lon = self.dict[key][0]
            lat = self.dict[key][1]
            dep = self.dict[key][2]
            mag = self.dict[key][3]
            relative_days = (self.dict[key][4]-reftime)/(24*60*60)
            _key = format(key,'8d')
            _lon = format(lon,'11.5f')
            _lat = format(lat,'10.5f')
            _dep = format(dep,'7.2f')
            _mag = format(mag,'4.1f')
            _relative_days = format(relative_days,'9.3f')
            line = _key+_lon+_lat+_dep+_mag+_relative_days
            f.write(line+"\n")
            if disp == True:
                print(line)
        print("Catalog information write into: ",info_file)
        f.close()

    def cata2fdsn(self,author="Hardy",catalog="SC",
                  cont="SC",contID="01",magtype="ML",
                  magauthor="SC Agency",elocname="SC",out_file='cata.fdsn'):

        cata2fdsn(self,author=author,catalog=catalog,                         
                  cont=cont,contID=contID,magtype=magtype,                       
                  magauthor=magauthor,elocname=elocname,out_file=out_file)

    def copy(self):
        return copy.deepcopy(self)
    def merge(self,dd2):
        for key in dd2.keys:
            if key in self.keys:
                raise Exception(f"Key error {key}. Please avoid using the same key value")
            self.dict[key]=dd2.dict[key]
        self.init()

    def __repr__(self):
        _qty = f"HypoDD relocation catalog with {len(self.dict.keys())} events\n"
        _time= f"     Time range is: {self.first_time+np.min(self.relative_seconds)} to {self.first_time+np.max(self.relative_seconds)}\n"
        _mag = f" Magnitue range is: {format(np.min(self.locs[:,3]),'4.1f')} to {format(np.max(self.locs[:,3]),'4.1f')}\n"
        _lon = f"Longitude range is: {format(np.min(self.locs[:,0]),'8.3f')} to {format(np.max(self.locs[:,0]),'8.3f')}\n"
        _lat = f" Latitude range is: {format(np.min(self.locs[:,1]),'7.3f')} to {format(np.max(self.locs[:,1]),'7.3f')}\n"
        _dep = f"    Depth range is: {format(np.min(self.locs[:,2]),'4.1f')} to {format(np.max(self.locs[:,2]),'4.1f')}\n"
        return _qty+_time+_mag+_lon+_lat+_dep
    
    def __getitem__(self,key):
        return self.dict[key]

def hypoinv2Catalog(inv):
    """
    Convert Hypoinv class to Catalog class
    """
    inv_dict={}
    for key in inv.dict_evid.keys():
        inv_dict[key] = inv.dict_evid[key][1:5]
        _time = inv.dict_evid[key][0]
        etime = UTCDateTime.strptime(_time,"%Y%m%d%H%M%S%f")
        inv_dict[key].append(etime)
    inv_cata = Catalog(locfile=None)
    inv_cata.dict = inv_dict
    inv_cata.init()
    return inv_cata



#-----------------------------------------------------------------------------
# coding: utf-8
# author: Hardy ZI
# history:
#		...
#       2021-03-15 Add CN2fsdn function
#

import re
from obspy import UTCDateTime
import os
import sys
import time
from seisloc.dd import loadDD
import time
from seisloc.geopara import WYpara
from tqdm import tqdm


def arc2cnv(arc_file,mag_threshold=-9,qty_limit=None):
    """
    convert y2000 archieve file into velest format file
    """
    count = 0
    cont = []    # Save the content of the phase file
    cont_sel = [] # content selected
    out_dict = {}     # output content
    with open(arc_file,'r') as f:
        for line in f:
            cont.append(line.rstrip())
    f.close()

    for line in tqdm(cont):
        if re.match("\d+",line[:6]):    # line start with digit is summary line
            e_label = line[:16]
            e_time = UTCDateTime.strptime(line[:12],'%Y%m%d%H%M')+int(line[12:16])*0.01
            e_lat = int(line[16:18])+int(line[19:23])*0.01/60
            e_lon = int(line[23:26])+int(line[27:31])*0.01/60
            e_dep = int(line[31:36])*0.01
            record_status = True
            try:
                if len(line) < 126:
                    e_mag = int(line[36:39])*0.01
                else:
                    e_mag = int(line[123:126])*0.01
            except:
                e_mag = 0
            if e_mag<mag_threshold:
                record_status=False
            if e_dep==0:                # no zero depth for VELEST
                record_status=False
            if qty_limit!=None and count >= qty_limit:
                record_status = False

            if record_status == False:
                continue
            
            cont_sel.append(line)
            count += 1
            out_dict[e_label] = {}    # dict with event line str as key
            out_dict[e_label]["e_time"]=e_time
            out_dict[e_label]["e_lat"]=e_lat
            out_dict[e_label]["e_lon"]=e_lon
            out_dict[e_label]["e_dep"]=e_dep
            out_dict[e_label]["e_mag"]=e_mag
            out_dict[e_label]["phase"]=[]   # array to store phases
        elif line[:6]=='      ':        # The last line indicate the evid
            if record_status==False:
                continue
            cont_sel.append(line)
            out_dict[e_label]["evid"]=int(line[63:72])
        else:                           # Station phase line
            if record_status==False:
                continue
            cont_sel.append(line)
            sta = line[:5].split()
            sta = sta[0]
            net = line[5:7]
            if line[14] == "P":
                pha = "P"
                phs_time = UTCDateTime.strptime(line[17:29],"%Y%m%d%H%M")+\
                            int(line[30:34])*0.01
                diff_time = phs_time - e_time
            elif line[47] == "S":
                pha = "S"
                phs_time = UTCDateTime.strptime(line[17:29],"%Y%m%d%H%M")+\
                            int(line[42:46])*0.01
                diff_time = phs_time - e_time
            pha_record = sta.ljust(5," ")+pha+"1"+format(diff_time,'6.2f')
            out_dict[e_label]["phase"].append(pha_record)
    print(f"# Total {count} events!")
    cnv_file = arc_file+".cnv"
    f = open(cnv_file,'w')
    for key in out_dict.keys():      # Loop for each event
        e_lat = out_dict[key]["e_lat"]
        e_lon = out_dict[key]["e_lon"]
        e_dep = out_dict[key]["e_dep"]
        e_mag = out_dict[key]["e_mag"]
        phases = out_dict[key]["phase"]
        evid = out_dict[key]["evid"]
        part1 = key[2:8]+" "+key[8:12]+" "+key[12:14]+"."+key[14:16]+" "
        part2 = format(e_lat,"7.4f")+"N"+" "+format(e_lon,"8.4f")+"E"+" "
        part3 = format(e_dep,"7.2f")+"  "+format(e_mag,"5.2f")
        
        f.write(part1+part2+part3)
        i = 0
        for phase in phases:
            if i%6==0:
                f.write("\n")
            f.write(phase)
            i=i+1
        tmp = i%6
        if tmp>0:
            f.write((6-tmp)*13*" "+"\n") # add space to fill line
        elif tmp==0:
            f.write("\n")
        f.write(f"   {str(evid).zfill(6)}\n")
    f.write("9999")              # indicates end of file for VELEST
    f.close()

    f = open(arc_file+".sel",'w')
    for line in cont_sel:
        f.write(line+"\n")
    f.close()

def dd2fdsn(in_file,subset=None):
    '''
    Convert the hypoDD reloc file into fdsn
    format that could be read by zmap	
    '''
    T0 = time.time()
    print("The start time is 0.")
    if subset == None:
        filt = False
    else:
        filt = True
        [lon_min,lon_max,lat_min,lat_max] = subset
    events = []
    out_file = "dd.fdsn"
    f=open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    f.close()
    eve_dict,df = loadDD(reloc_file=in_file)
    T1 = time.time()
    print("%f seconds passed to load hypoDD file" %(T1-T0))
    eve_list = list(eve_dict)
    f = open(out_file,'a')
    for eve in eve_list:
        evid = eve_dict[eve][4]
        e_time = UTCDateTime.strptime(eve,"%Y%m%d%H%M%s%f")
        e_lon = eve_dict[eve][0]
        e_lat = eve_dict[eve][1]
        e_dep = eve_dict[eve][2]
        e_mag = eve_dict[eve][3] 
        if filt:
            if e_lat>lat_max or e_lat<lat_min or e_lon>lon_max or e_lon<lon_min:
                continue
        mag_type = 'ML'
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(format(e_lat,'6.3f')+"|")
        f.write(format(e_lon,'7.3f')+"|")
        f.write(format(e_dep,'6.2f')+"|")
        f.write("Hardy|")
        f.write("SC|")
        f.write("SC|")
        f.write("01|")
        f.write(mag_type+'|')
        f.write(format(e_mag+0.01,'5.2f')+"|")
        f.write("SC Agency|")
        f.write("SC\n")
    f.close()

def sum2fdsn(in_file,subset=None):
    '''
    Convert the out.sum file into fdsn
    format that could be read by zmap	
    '''
    if subset == None:
        filt = False
    else:
        filt = True
        [lon_min,lon_max,lat_min,lat_max] = subset
    events = []
    if isinstance(in_file,list):
        for file in in_file: #input file, *.csv, without two header line
            with open(file,'r',encoding='UTF-8') as f:
                for line in f:
                    events.append(line.rstrip())
            f.close()
    else:
        with open(in_file,'r',encoding='UTF-8') as f:
            for line in f:
                events.append(line.rstrip())
        f.close()
    out_file = "sum.fdsn"
    f=open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    f.close()
    f = open(out_file,'a')
    for event in events:
        evid = int(event[141:146])
        e_time = UTCDateTime(event[:16],"%Y%m%d%H%M%S%f")
        e_lat = int(event[16:18])+int(event[19:23])/100.0/60.0
        e_lon = int(event[23:26]) + int(event[27:31])/100.0/60.0
        if filt:
            if e_lat>lat_max or e_lat<lat_min or e_lon>lon_max or e_lon<lon_min:
                continue
        e_dep = int(event[31:36])/100.0
        e_mag = int(event[123:126])/100.0
        mag_type = 'ML'
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(format(e_lat,'6.3f')+"|")
        f.write(format(e_lon,'7.3f')+"|")
        f.write(format(e_dep,'6.2f')+"|")
        f.write("Hardy|")
        f.write("SC|")
        f.write("SC|")
        f.write("01|")
        f.write(mag_type+'|')
        f.write(format(e_mag+0.01,'5.2f')+"|")
        f.write("SC Agency|")
        f.write("SC\n")
    f.close()


def SC2fdsn(in_file,subset=None):
    '''
    Convert the event file from the SC catalog into fdsn
    format that could be read by zmap	
    '''
    if subset == None:
        filt = False
    else:
        filt = True
        [lon_min,lon_max,lat_min,lat_max] = subset
    events = []
    if isinstance(in_file,list):
        for file in in_file: #input file, *.csv, without two header line
            with open(file,'r',encoding='UTF-8') as f:
                for line in f:
                    events.append(line.rstrip())
            f.close()
    else:
        with open(in_file,'r',encoding='UTF-8') as f:
            for line in f:
                events.append(line.rstrip())
        f.close()
    out_file = "out.fdsn"
    f=open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    f.close()
    f = open(out_file,'a')
    for event in events:
        evid = int(event[:5])
        e_time = UTCDateTime(event[6:27])
        e_lat = event[28:34]
        e_lon = event[35:42]
        if filt:
            if float(e_lat)>lat_max or float(e_lat)<lat_min or float(e_lon)>lon_max or float(e_lon)<lon_min:
                continue
        e_dep = str(int(int(event[47:])/1000))
        e_mag = float(event[43:46])+0.01
        mag_type = 'ML'
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(e_lat+"|")
        f.write(e_lon+"|")
        f.write(e_dep+"|")
        f.write("Hardy|")
        f.write("SC|")
        f.write("SC|")
        f.write("01|")
        f.write(mag_type+'|')
        f.write(format(e_mag+0.01,'5.2f')+"|")
        f.write("SC Agency|")
        f.write("SC\n")
    f.close()

def CN2fdsn(in_file):
    '''
    Convert the event file from the China National Data Center into fdsn
    format that could be read by zmap	
    '''
    events = []
    if isinstance(in_file,list):
        for file in in_file: #input file, *.csv, without two header line
            with open(file,'r',encoding='UTF-8-sig') as f:
                for line in f:
                    events.append(line.rstrip())
            f.close()
    else:
        with open(in_file,'r',encoding='UTF-8-sig') as f:
            for line in f:
                events.append(line.rstrip())
        f.close()
    out_file = "out.fdsn"
    f=open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    f.close()
    evid = 0
    f = open(out_file,'a')
    for event in events:
        evid = evid+1
        e_time,e_lon,e_lat,e_dep,mag_type,e_mag,_,_ = re.split(",",event)
        e_date,e_hm = re.split(" ",e_time)
        e_year,e_month,e_day = re.split("/",e_date)
        e_year=int(e_year)
        e_month = int(e_month)
        e_day = int(e_day)
        e_hr,e_min = re.split(":",e_hm)
        e_hr = int(e_hr)
        e_min = int(e_min)
        e_time = UTCDateTime(e_year,e_month,e_day,e_hr,e_min,0,0)
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(e_lat+"|")
        f.write(e_lon+"|")
        f.write(e_dep+"|")
        f.write("Hardy|")
        f.write("SC|")
        f.write("SC|")
        f.write("01|")
        f.write(mag_type+'|')
        f.write(format(float(e_mag)+0.01,'5.2f')+"|")
        f.write("SC Agency|")
        f.write("SC\n")
    f.close()

def IRIS2fdsn(events):
    '''
    Convert the event file from the python client into fdsn
    format that could be read by zmap	
    '''
    out_file = "out.fdsn"
    f=open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    f.close()
    evid = 0
    f = open(out_file,'a')
    for event in events:
        evid = evid+1
        e_time = event["origins"][0]["time"]
        e_lon = event["origins"][0]["longitude"]
        e_lat = event["origins"][0]["latitude"]
        e_mag = event["magnitudes"][0]["mag"]
        try:
            e_dep = event["origins"][0]["depth"]/1000
        except:
            e_dep = 0
        try:
            mag_type = event["magnitudes"][0]["magnitude_type"]
        except:
            mag_type = None
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(str(e_lat)+"|")
        f.write(str(e_lon)+"|")
        f.write(str(e_dep)+"|")
        f.write("Hardy|")
        f.write("IRIS|")
        f.write("IRIS|")
        f.write("01|")
        f.write(str(mag_type)+'|')
        f.write(format(float(e_mag)+0.01,'5.2f')+"|")
        f.write("SC Agency|")
        f.write("SC\n")
    f.close()

def cata2fdsn(cata,author="Hardy",catalog="SC",
              cont="SC",contID="01",magtype="ML",
              magauthor="SC Agency",elocname="SC",out_file='cata.fdsn'):
    """
    Outout fdsn format to be read by ZMAP
    """
    keys = cata.keys
    f = open(out_file,'w')
    f.write("#EventID|Time|Latitude|Longitude|Depth/km|Author|Catalog"
            +"|Contributor|ContributorID|MagType|Magnitude|MagAuthor|EventLocationName\n")
    for evid in keys:
        e_time = cata[evid][4]
        e_lon = cata[evid][0]
        e_lat = cata[evid][1]
        e_dep = cata[evid][2]
        e_mag = cata[evid][3]
        magtype = 'ML'
        f.write('{:0>6d}'.format(evid)+"|")
        f.write(str(e_time)+"|")
        f.write(format(e_lat,'6.3f')+"|")
        f.write(format(e_lon,'7.3f')+"|")
        f.write(format(e_dep,'6.2f')+"|")
        f.write(author+"|")
        f.write(catalog+"|")
        f.write(cont+"|")
        f.write(contID+"|")
        f.write(magtype+'|')                                                                                                                                                                               
        f.write(format(e_mag+0.01,'5.2f')+"|")
        f.write(magauthor+"|")
        f.write(elocname+"\n")
    f.close()

def ncsn2pha(source_file,target_file):
    input_content=[]
    output_content=[]
    with open(source_file,"r") as f:
        for line in f:
            input_content.append(line.rstrip())
    f.close()
    for line in input_content:
        if line[0:7]=="       ": # Pass event id line
            continue
        if re.match("\d+",line[0:7]): # A event line
            date = line[0:8]
            mo = line[5:6]
            dy = line[6:8]
            hr = line[8:10]
            min = line[10:12]
            sec = line[12:16]
            deglat = line[16:18]
            minlat = line[19:23]
            deglon = line[23:26]
            minlon = line[27:31]
            depth = int(line[31:36])
            mag = line[147:150]
            res = line[48:52]
            herr = line[85:89]
            verr = line[89:93]
            cuspid = line[136:146]

            year = line[0:4]
            lat = int(deglat)+int(minlat)*0.01/60
            lon = int(deglon)+int(minlon)*0.01/60

            part1 = "# "+year+str(int(mo)).rjust(3," ")+str(int(dy)).rjust(3," ")
            part2 = str(int(hr)).rjust(3," ")+str(int(min)).rjust(3," ")+format(int(sec)*0.01,'6.2f')+'  '
            part3 = format(lat,"7.4f")+"  "+format(lon,"8.4f")+"   "+format(depth*0.01,'5.2f')+"  "+format(int(mag)*0.01,'4.2f')
            part4 = " "+format(int(herr)*0.01,'5.2f')+" "+format(int(verr)*0.01,'5.2f')+" "+format(int(res)*0.01,'5.2f')+" "+cuspid
            output_content.append(part1+part2+part3+part4)
        else:
            c1 = line[3:4]
            c2 = line[9:10]
            c4 = line[11:12]
            premk = line[13:15]
            sremk = line[46:48]
            sta = line[5:7]+line[0:5]
            pqual = int(line[16:17])
            squal = int(line[49:50])
            p_min = line[27:29]
            p_sec = line[29:34]
            s_sec = line[41:46]
            p_res = int(line[34:39])*0.01
            s_res = int(line[50:54])*0.01
            p_imp = line[100:104]
            try:
                p_imp = int(p_imp)*0.001
            except:
                continue
            s_imp = line[104:108]
            try:
                s_imp = int(s_imp)*0.001
            except:
                continue
            if pqual < 4 and premk != "  " and c4 == "Z":
                if p_res > 0.2:
                    continue
                if pqual == 0:
                    p_weight = 1.0
                elif pqual == 1:
                    p_weight = 0.5
                elif pqual == 2:
                    p_weight = 0.2
                elif pqual == 3:
                    p_weight = 0.1
                else:
                    p_weight = 0.0

                if p_imp > 0.5:
                    p_weight = -1*p_weight
                if int(p_min) < int(min): #60 -> 01 of another hour
                    dmin = int(p_min) + 60 - int(min)
                else:
                    dmin = int(p_min) - int(min)
                p_time = dmin*60 + (int(p_sec)-int(sec))*0.01
                output_content.append(sta+"    "+format(p_time,'6.3f')+format(p_weight,'8.3f')+"   P")

            if squal < 4 and sremk != "  " and c4 == "Z":
                if s_res>0.2:
                    continue
                if squal == 0:
                    s_weight = 1.0
                elif squal == 1:
                    s_weight = 0.5
                elif squal == 2:
                    s_weight = 0.2
                elif squal == 3:
                    s_weight = 0.1
                else:
                    s_weight = 0.0

                if s_imp > 0.5:
                    s_weight = -1*s_weight
                if int(p_min) < int(min): #60 -> 01 of another hour
                    dmin = int(p_min) + 60 - int(min)
                else:
                    dmin = int(p_min) - int(min)
                s_time = dmin*60 + (int(s_sec)-int(sec))*0.01
                output_content.append(sta+"    "+format(s_time,'6.3f')+format(s_weight,'8.3f')+"   S")
    with open(target_file,"w") as f:
        for line in output_content:
            f.write(line)
            f.write('\n')
                
def sc2phs(file_list=[],trims=[None,None,None,None],magThreshold=None,baseid=1,outfile='out.phs'):
    """
    convert Sichuan earthquake administration report into file could be
    recognized by Hypoinverse.
    """
    if file_list == []:
        for file in os.listdir("./"):
            if file[-4:]==".adj":
                file_list.append(file)
    file_list.sort()
 
    lon_filt=True
    lat_filt=True
    mag_filt=True #filt magnitude
    time_filt=True
    lon_min,lon_max,lat_min,lat_max= trims
    if lon_min==None:
        lon_filt=False
    if lat_min==None:
        lat_filt=False
    if magThreshold==None:
        mag_filt = False
    #initiate part
    output_content=[]
    input_content = []
    for file in file_list:
        with open(file,"r") as f:
            for line in f:
                input_content.append(line.rstrip())
        f.close()
    evid= baseid-1
    for line in tqdm(input_content):
        if re.match("\d+",line[3:7]) and line[7]=="/":#then it is a event line
            record_status=True
            e_year = line[3:7]
            e_month = line[8:10]
            e_day = line[11:13]
            e_hour = line[14:16]
            e_minute = line[17:19]
            e_second_int = line[20:22]
            e_second_left = line[23:24]
            e_lat=line[26:32]
            if e_lat=='      ':
                record_status=False
                continue
            else:
                e_lat=float(e_lat)

                if lat_filt:      
                    if e_lat<lat_min or e_lat>lat_max:
                        record_status=False
                        continue  
            e_lon=line[34:41]     
            if e_lon=='       ':  
                record_status=False      
                continue          
            else:                 
                e_lon=float(e_lon)
                if lon_filt:      
                    if e_lon<lon_min or e_lon>lon_max:
                        record_status=False
                        continue  
            e_dep=line[43:45]     
            if e_lat=='  ':       
                record_status=False      
                continue          
            else:                 
                e_lat=float(e_lat)
            e_mag=line[47:50]     
            if e_mag=='   ':      
                record_status=False      
                continue          
            else:                 
                e_mag=float(e_mag)
                if mag_filt:      
                    if e_mag < magThreshold:
                        record_status=False
                        continue  
            if record_status==True:      
                if evid!=baseid-1:   
                    output_content.append(format(str(evid).rjust(72," ")))
                evid+=1     
                part1 = e_year+e_month+e_day+e_hour+e_minute+e_second_int+e_second_left+'0'
                part2 = str(int(e_lat))+" "+str(int((e_lat-int(e_lat))*60*100)).zfill(4)
                part3 = str(int(e_lon))+"E"+str(int((e_lon-int(e_lon))*60*100)).zfill(4)
                part4 = str(int(e_dep)*100).rjust(5," ")+"000"+"L".rjust(84," ")+str(int(e_mag*100)).zfill(3)
                output_content.append(part1+part2+part3+part4)
                                  
        elif record_status==True: 
            if line[0:2]!="  ":   
                net = line[0:2]   
                sta = re.split(" +",line[3:8])[0]
            p_type = line[17:19]  
            p_weight = float(line[25:28])
            if p_weight >= 0.75:  
                weightCode = 1    
            elif p_weight >= 0.5: 
                weightCode = 2    
            elif p_weight >= 0.0: 
                weightCode = 3    
            p_hour = line[32:34]  
            p_minute = line[35:37]
            p_seconds = float(line[38:43])
            p_residual=line[45:50]
            if p_residual=="     ":
                continue                     
            else:         
                try:      
                    p_residual = float(p_residual)
                except:   
                    continue  
            if p_type =="Pg": 
                part1 = sta.ljust(5," ")+net+"  SHZ IPU"+str(weightCode)+e_year+e_month+e_day+p_hour+p_minute+" "+str(int(p_seconds*100)).zfill(4)
                part2 = str(int(p_residual*100)).rjust(4," ")+"  0    0   0   0"
                output_content.append(part1+part2)
            elif p_type == "Sg": 
                part1 = sta.ljust(5," ")+net+"  SHZ     "+e_year+e_month+e_day+p_hour+p_minute+"    0   0  0 "
                part2 = str(int(p_seconds*100)).zfill(4)+"ES "+str(weightCode)+str(int(p_residual*100)).rjust(4," ")
                output_content.append(part1+part2) 
    output_content.append(format(str(evid).rjust(72," ")))
    with open(outfile,"w") as f:
        for line in output_content:
            f.write(line+"\n")

def getWeightCode(weight,stds):
    """
    Get the corresponding weight code as HYPOINVERSE
    """
    if weight < stds[3]:
        weightCode = 4
    elif weight < stds[2]:
        weightCode = 3
    elif weight < stds[1]:
        weightCode = 2
    elif weight < stds[0]:
        weightCode = 1
    else:
        weightCode = 0
        
    return weightCode

def real2arc(inFile,
             minObs=8,
             cmp="SHZ",
             boundCond=[],
             weightCodeStds=[0.95,0.75,0.5,0.25],
             startId = 0):
    """
    change from REAL association result to the file that could be read by HYPOINVERSE
    outfile have the same name of input file with suffix '.phs'
    
    Parameters:
        inFile: input file generated by REAL
        minObs: minimum observation by stations, including P and S
        boundCond: boundary condition, [lonMin,lonMax,latMin,latMax]
        weightCodeStds: standards for weight code assignment,corresponding weight codes
                        are [0,1,2,3,4]
        startId: Initial event id. Default is zero, and events will receive IDs start from startID+1.
                 Under cases user have demand to merge different catalogs, it is necessary to set up 
                 different event ids to avoid event id confliction
        
    """
    if len(boundCond) == 0:
        boundFilt=False
    elif len(boundCond) == 4:
        boundFilt=True
        lonMin,lonMax,latMin,latMax = boundCond
    else:
        raise Exception("boundCond should be with length 0 or 4!")
        
    inTitle = re.split("\.",inFile)[:-1]
    inTitle = "".join(inTitle)
    outFile = inTitle+".phs"

    cont=[]
    with open(inFile,"r") as f:
        for line in f:
            cont.append(line.rstrip())
    f.close()
    loopId = startId                          # initiate event ID
    f = open(outFile,"w")
    for line in cont:
        f_para = re.split(" +",line)[1]       # first parameter            
        if re.match("\d+",f_para):            # event line start with digital number
            status = False                    # initiate the status to be False
            _,_no,_yr,_mo,_dy,_otime,_absec,_timeRes = re.split(" +",line)[:8]
            _lat,_lon,_dep,_mag,_magRes = re.split(" +",line)[8:13]
            _numP,_numS,_numT,_staGap = re.split(" +",line)[13:]
            lon = float(_lon); lat = float(_lat);dep = float(_dep)
            numT = int(_numT)
            if boundFilt:
                if lon<lonMin or lon>lonMax or lat<latMin or lat>latMax:
                    continue
            if numT < minObs:
                continue
            status = True
            # REAL provide seconds below zero and above 60, need to handle it
            _ehr,_emin,_esec=re.split(":",_otime)
            esec = float(_esec)
            etime = UTCDateTime(_yr+"-"+_mo+"-"+_dy+"T"+_ehr+":"+_emin+":"+"00")+esec
            # re-get the value
            eyear = etime.year
            emonth = etime.month
            eday = etime.day
            ehr = etime.hour
            emin = etime.minute
            esec = etime.second+etime.microsecond/1000000

            latInt = int(lat)             # int part, unit degree
            latDec = lat - latInt         # float part, unit degree
            lonInt = int(lon)             # int part, unit degree
            lonDec = lon - lonInt         # float part, unit degree

            if _mag == "-inf":
                mag = 0.0
                mag_res = 0.0
            else:
                mag = float(_mag)
                mag_res = float(_magRes)
            if loopId > startId:
                f.write(format(loopId,">72d")+"\n")
            f.write(format(eyear,"4d")+format(emonth,"0>2d")+format(eday,"0>2d")+\
                format(ehr,"0>2d")+format(emin,"0>2d")+format(esec*100,"0>4.0f")+\
                format(latInt,"0>2d")+" "+format(latDec*60*100,"0>4.0f")+\
                format(lonInt,"0>3d")+"E"+format(lonDec*60*100,"0>4.0f")+\
                format(dep*100,">5.0f")+format(mag*100,"0>3.0f")+"\n")                
            loopId = loopId + 1
        else:
            if status == False:
                continue
            _,net,sta,phaType,_absSec,_relSec,_amp,_res,_weight,_az = re.split(" +",line)
            res=float(_res);weight=float(_weight)
            weightCode = getWeightCode(weight,weightCodeStds)
            phaTime = etime + float(_relSec)
            phaYr = phaTime.year
            phaMo = phaTime.month
            phaDy = phaTime.day
            phaHr=phaTime.hour
            phaMin=phaTime.minute
            phaSec=phaTime.second+phaTime.microsecond/1000000
            phaSecInt = phaTime.second #int part
            phaSecDec = phaTime.microsecond/1000000 #float part
            #write in content
            if phaType=="P":
                f.write(format(sta,"<5s")+format(net,"2s")+"  "+format(cmp,'3s')+\
                    " IPU"+str(weightCode)+\
                    format(phaYr,"4d")+format(phaMo,"0>2d")+format(phaDy,"0>2d")+\
                    format(phaHr,"0>2d")+format(phaMin,"0>2d")+\
                    format(phaSec*100,">5.0f")+\
                    format(res*100,">4.0f")+"  0    0   0   0"+"\n")
            if phaType=="S":
                f.write(format(sta,"<5s")+format(net,"2s")+"  "+format(cmp,'3s')+"     "+\
                    format(phaYr,"4d")+format(phaMo,"0>2d")+format(phaDy,"0>2d")+\
                    format(phaHr,"0>2d")+format(phaMin,"0>2d")+"    0   0  0"+\
                    format(phaSec*100,">5.0f")+"ES "+str(weightCode)+format(res*100,">4.0f")+"\n")
    if loopId != startId:
        f.write(format(loopId,">72d")+"\n")
    f.close()



import matplotlib.pyplot as plt
from obspy import UTCDateTime
import numpy as np
from seisloc.utils import time_interval_ticks

def segmented_plot(xys,base_time,secs,tick_secs,
                   xlim=[],
                   ylim=[],
                   columns=4,
                   subplotsize=(3,3),
                   marker='o',
                   ms=1,
                   unit='month',
                   title_fontsize=12,
                   wspace=None,
                   hspace=None,
                   aspect='equal'):
    """
    Description
    Do subplots for xys by time intervals.
    
    Parameters
           xys: 2-D numpy array for location
     base_time: base UTCDateTime time, 'secs' and 'tick_secs' are relative seconds with reference to 'base_time'
     xlim,ylim: plot range, min(x),max(x),min(y),max(y) used if not set
       columns: columns for each row
  subplotsize: figure size for each subplot
     marker,ms: marker and markersize
          unit: the interval type for each subplot, it could be 'year','month','day','hour','minute', or 'second', 
                this parameter is designed to control subplot title format.
     wspace,hspace,aspect: refer to matplotlib.pyplot for details
     
     Return:
           axs: one dimensional list of axes
    """
    # format control
    assert len(xys.shape) == 2
    assert isinstance(base_time,UTCDateTime)
    assert len(secs.shape) == 1
    assert len(tick_secs.shape) == 1
    
    if xlim == []:
        xlim = [np.min(xys[:,0]),np.max(xys[:,0])]
    if ylim == []:
        ylim = [np.min(xys[:,1]),np.max(xys[:,1])]

    # set up subplots
    segment_qty = len(tick_secs) - 1
    

    if segment_qty <= columns:
        rows = 1
        columns = segment_qty
    else:
        if segment_qty % columns == 0:
            rows = int(segment_qty/columns)
        else:
            rows = int(segment_qty/columns)+1
    fig,axs = plt.subplots(rows,columns,
                           sharex=True,sharey=True,
                          figsize=(columns*subplotsize[0],rows*subplotsize[1]))
    if len(axs)>1:
        axs = axs.ravel()

    # title format
    if unit == 'year':
        title_fmt = "%Y"
    elif unit == 'month':
        title_fmt = "%Y-%m"
    elif unit == 'day':
        title_fmt = "%Y-%m-%d"
    elif unit == 'hour':
        title_fmt = "%Y-%m-%dT%H"
    elif unit == 'minute':
        title_fmt = "%Y-%m-%dT%H:%M"
    elif unit == 'second':
        title_fmt = "%Y-%m-%dT%H:%M:%S"
    else:
        raise Exception("Wrong unit type, should be 'year','month','day','hour','minute','second'")
    
    # subplots plot in here
    for i in range(segment_qty):
        plt.sca(axs[i])
        plt.xlim(xlim)
        plt.ylim(ylim)
        kk = np.where((secs>=tick_secs[i])&(secs<tick_secs[i+1]))
        loop_time = base_time + tick_secs[i]
        loop_yr = loop_time.year
        loop_mo = loop_time.month
        loop_dy = loop_time.day
        loop_hr = loop_time.hour
        loop_min = loop_time.minute
        loop_sec = loop_time.second
        plt.plot(xys[kk,0],xys[kk,1],marker=marker,c='k',mfc='none',ms=ms,zorder=10)
        
        plt.title(loop_time.strftime(title_fmt),fontsize=title_fontsize)

        plt.gca().set_aspect(aspect)
        
    for i in range(segment_qty,len(axs)):
        plt.sca(axs[i])               # set current axis
        plt.axis('off')
        
    # adjust space    
    if wspace != None or hspace != None:
        plt.subplots_adjust(wspace=wspace,hspace=hspace)
    else:
        plt.tight_layout()
    
    return axs

def intervals_plot(xys,
                rela_secs,
                reftime,
                interval=1,
                method='month',
                xlim=[],
                ylim=[],
                columns=4,
                subplotsize=(3,3),
                marker='o',
                ms=1,
                wspace=None,
                hspace=None):
    """
    Description
        Plot xy location subplots by intervals
        
    Parameters
    """
    secs = rela_secs.copy()
    
    # format control
    assert len(xys.shape) == 2
    assert len(secs.shape) == 1
    
    if xlim == []:
        xlim = [np.min(xys[:,0]),np.max(xys[:,0])]
    if ylim == []:
        ylim = [np.min(xys[:,1]),np.max(xys[:,1])]
    min_time = reftime + np.min(secs)
    max_time = reftime + np.max(secs)
    base_time, tick_secs = time_interval_ticks(min_time,max_time,interval=interval,unit=method)
    
    diff_time = reftime - base_time
    secs -= diff_time
    
    axs = segmented_plot(xys,base_time,secs,tick_secs,
                         xlim=xlim,ylim=ylim,
                         marker=marker,ms=ms,
                         columns=columns,
                         unit=method,
                         subplotsize=subplotsize,
                         wspace=wspace,
                         hspace=hspace)
    
    return axs

def depths_plot(xyz,
                deplim=[0,10],interval=1,
                xlim=[],ylim=[],
                columns=4,subplotsize=(3,3),
                marker='o',ms=1,color='k',
                zorder=0,
                wspace=None,hspace=None):
    
    assert len(xyz.shape) == 2
    
    if xlim == []:
        xlim = [np.min(xyz[:,0]),np.max(xyz[:,0])]
    if ylim == []:
        ylim = [np.min(xyz[:,1]),np.max(xyz[:,1])]

    depticks = list(range(deplim[0],deplim[1],interval))
    depticks.append(deplim[-1])
    
    subqty = len(depticks) - 1
    if subqty <= columns:
        rows = 1
        columns = subqty
    else:
        if subqty % columns == 0:
            rows = int(subqty/columns)
        else:
            rows = int(subqty/columns)+1
    fig,axs = plt.subplots(rows,columns,
                           sharex=True,
                           sharey=True,
                           figsize=(columns*subplotsize[0],rows*subplotsize[1]))
    axs = axs.ravel()
    
    for i in range(subqty):
        plt.sca(axs[i])
        plt.xlim(xlim)
        plt.ylim(ylim)
        deplow = depticks[i]
        dephigh = depticks[i+1]
        kk = np.where((xyz[:,2]>deplow)&(xyz[:,2]<dephigh))
        plt.plot(xyz[kk,0],xyz[kk,1],
                 marker=marker,c=color,ms=ms,
                 zorder=zorder)
        plt.title(f"{deplow}-{dephigh} km")
        plt.gca().set_aspect("equal")
        
    for i in range(subqty,len(axs)):
        plt.sca(axs[i])
        plt.axis("off")

    return axs
